
import pdb
from abc import ABC, abstractmethod
import torch
# import warnings

class ExpectedLogLikelihood(ABC):

    def __init__(self, svEmbeddingAllTimes, linkFunction):
        self._svEmbeddingAllTimes = svEmbeddingAllTimes
        self._linkFunction = linkFunction

    def get_svEmbeddingAllTimes(self):
        return self._svEmbeddingAllTimes

    @abstractmethod
    def evalSumAcrossTrialsAndNeurons(self, svPosteriorOnLatentsStats=None):
        pass

    @abstractmethod
    def buildKernelsMatrices(self):
        pass

    @abstractmethod
    def computeSVPosteriorOnLatentsStats(self):
        pass

    @abstractmethod
    def setMeasurements(self, measurements):
        pass

    @abstractmethod
    def setIndPointsLocs(self, locs):
        pass

    @abstractmethod
    def setKernels(self, kernels):
        pass

    @abstractmethod
    def setInitialParams(self, initial_params):
        pass

    @abstractmethod
    def setELLCalculationParams(self, eLLCalculationParams):
        pass

    def sampleCIFs(self, times, nudget=1e-3):
        h = self._svEmbeddingAllTimes.sample(times=times, nudget=nudget)
        nTrials = len(h)
        answer = [self._linkFunction(h[r]) for r in range(nTrials)]
        return answer

#     def computeCIFsMeans(self, times):
#         # h \in nTrials x times x nNeurons
#         h = self._svEmbeddingAllTimes.computeMeans(times=times)
#         nTrials = h.shape[0]
#         nNeurons = h.shape[2]
#         answer = [[self._linkFunction(h[r,:,n]) for n in range(nNeurons)] for r in range(nTrials)]
#         return answer
# 
    def computeExpectedPosteriorCIFs(self, times):
        # h \in nTrials x times x nNeurons
        # eMean, eVar = self._svEmbeddingAllTimes.computeMeansAndVars(times=times)
        # answer = self._getELinkValues(eMean=eMean, eVar=eVar)
        # return answer

        eMean, eVar = self._svEmbeddingAllTimes.predict(times=times)
        nTrials = eMean.shape[0]
        nNeurons = eMean.shape[2]
        answer = [[self._linkFunction(eMean[r, :, n]+0.5*eVar[r, :, n])
                   for n in range(nNeurons)] for r in range(nTrials)]
        return answer

    def getSVPosteriorOnIndPointsParams(self):
        return self._svEmbeddingAllTimes.getSVPosteriorOnIndPointsParams()

    def getSVEmbeddingParams(self):
        return self._svEmbeddingAllTimes.getParams()

    def computeEmbeddingsMeansAndVarsAtTimes(self, times):
        return self._svEmbeddingAllTimes.computeMeansAndVarsAtTimes(times)

    def getIndPointsLocs(self):
        return self._svEmbeddingAllTimes.getIndPointsLocs()

    def getKernels(self):
        return self._svEmbeddingAllTimes.getKernels()

    def getKernelsParams(self):
        return self._svEmbeddingAllTimes.getKernelsParams()

    def predictLatents(self, times):
        return self._svEmbeddingAllTimes.predictLatents(times=times)

    def predictEmbedding(self, times):
        return self._svEmbeddingAllTimes.predict(times=times)

    def setPriorCovRegParam(self, priorCovRegParam):
        self._svEmbeddingAllTimes.setPriorCovRegParam(priorCovRegParam=priorCovRegParam)

class PointProcessELL(ExpectedLogLikelihood):
    def __init__(self, svEmbeddingAllTimes, svEmbeddingAssocTimes, linkFunction):
        super().__init__(svEmbeddingAllTimes=svEmbeddingAllTimes, linkFunction=linkFunction)
        self._svEmbeddingAssocTimes = svEmbeddingAssocTimes

    def evalSumAcrossTrialsAndNeurons(self, svPosteriorOnLatentsStats=None):
        if svPosteriorOnLatentsStats is not None:
            svPosteriorOnLatentsStatsAllTimes = \
             svPosteriorOnLatentsStats["allTimes"]
            svPosteriorOnLatentsStatsAssocTimes = \
             svPosteriorOnLatentsStats["assocTimes"]
        else:
            svPosteriorOnLatentsStatsAllTimes = None
            svPosteriorOnLatentsStatsAssocTimes = None
        eMeanAllTimes, eVarAllTimes = \
            self._svEmbeddingAllTimes.computeMeansAndVars(
                svPosteriorOnLatentsStats=svPosteriorOnLatentsStatsAllTimes)
        eMeanAssocTimes, eVarAssocTimes = \
            self._svEmbeddingAssocTimes.computeMeansAndVars(
                svPosteriorOnLatentsStats=svPosteriorOnLatentsStatsAssocTimes)
        eLinkValues = self._getELinkValues(eMean=eMeanAllTimes,
                                           eVar=eVarAllTimes)
        eLogLinkValues = self._getELogLinkValues(eMean=eMeanAssocTimes,
                                                 eVar=eVarAssocTimes)
        # self._legQuadWeights \in nTrials x nQuadHerm x 1
        # aux0 \in nTrials x 1 x nQuadHerm
        aux0 = torch.transpose(input=self._legQuadWeights, dim0=1, dim1=2)
        # eLinkValues \in  nTrials x nQuadHerm x nNeurons
        # aux1 \in  nTrials x 1 x nNeurons
        aux1 = torch.matmul(aux0, eLinkValues)
        sELLTerm1 = torch.sum(aux1)
        sELLTerm2 = torch.sum(eLogLinkValues)
        answer = -sELLTerm1+sELLTerm2
        return answer

    def buildKernelsMatrices(self):
        self._svEmbeddingAllTimes.buildKernelsMatrices()
        self._svEmbeddingAssocTimes.buildKernelsMatrices()

    def computeSVPosteriorOnLatentsStats(self):
        allTimesStats = self._svEmbeddingAllTimes.\
            computeSVPosteriorOnLatentsStats()
        assocTimesStats = self._svEmbeddingAssocTimes.\
            computeSVPosteriorOnLatentsStats()
        answer = {"allTimes": allTimesStats, "assocTimes": assocTimesStats}
        return answer

    def setMeasurements(self, measurements):

        stackedSpikeTimes, neuronForSpikeIndex = \
            self.__stackSpikeTimes(spikeTimes=measurements)
        self._svEmbeddingAssocTimes.setTimes(times=stackedSpikeTimes)
        self._svEmbeddingAssocTimes.setNeuronForSpikeIndex(neuronForSpikeIndex=
                                                            neuronForSpikeIndex)

    def __stackSpikeTimes(self, spikeTimes):
        # spikeTimes list[nTrials][nNeurons][nSpikes]
        nTrials = len(spikeTimes)
        stackedSpikeTimes = [[] for i in range(nTrials)]
        neuronForSpikeIndex = [[] for i in range(nTrials)]
        for trialIndex in range(nTrials):
            aList = [spikeTime
                     for neuronIndex in range(len(spikeTimes[trialIndex]))
                     for spikeTime in spikeTimes[trialIndex][neuronIndex]]
            stackedSpikeTimes[trialIndex] = torch.tensor(aList)
            # stackedSpikeTimes[trialIndex] = torch.unsqueeze(
            #     stackedSpikeTimes[trialIndex], 1)
            aList = [neuronIndex
                     for neuronIndex in range(len(spikeTimes[trialIndex]))
                     for spikeTime in spikeTimes[trialIndex][neuronIndex]]
            neuronForSpikeIndex[trialIndex] = torch.tensor(aList)
        return stackedSpikeTimes, neuronForSpikeIndex

    def setIndPointsLocs(self, locs):
        self._svEmbeddingAllTimes.setIndPointsLocs(locs=locs)
        self._svEmbeddingAssocTimes.setIndPointsLocs(locs=locs)

    def setKernels(self, kernels):
        self._svEmbeddingAllTimes.setKernels(kernels=kernels)
        self._svEmbeddingAssocTimes.setKernels(kernels=kernels)

    def setInitialParams(self, initial_params):
        self._svEmbeddingAllTimes.setInitialParams(initial_params=initial_params)
        self._svEmbeddingAssocTimes.setInitialParams(initial_params=initial_params)

    def setELLCalculationParams(self, eLLCalculationParams):
        self._svEmbeddingAllTimes.setTimes(times=eLLCalculationParams["leg_quad_points"])
        self._legQuadWeights = eLLCalculationParams["leg_quad_weights"]

    @abstractmethod
    def _getELinkValues(self, eMean, eVar):
        pass

    @abstractmethod
    def _getELogLinkValues(self, eMean, eVar):
        pass

class PointProcessELLExpLink(PointProcessELL):
    def __init__(self, svEmbeddingAllTimes, svEmbeddingAssocTimes):
        super().__init__(svEmbeddingAllTimes=svEmbeddingAllTimes,
                         svEmbeddingAssocTimes=svEmbeddingAssocTimes,
                         linkFunction=torch.exp)

    def _getELinkValues(self, eMean, eVar):
        # eLinkValues \in nTrials x nQuadLeg x nNeurons
        eLinkValues = self._linkFunction(input=eMean+0.5*eVar)
        return eLinkValues

    def _getELogLinkValues(self, eMean, eVar):
        # eLogLink = torch.cat([torch.squeeze(input=eMean[trial]) for trial in range(len(eMean))])
        eLogLink = torch.cat([eMean[trial] for trial in range(len(eMean))])
        return eLogLink

class PointProcessELLQuad(PointProcessELL):

    def __init__(self, svEmbeddingAllTimes, svEmbeddingAssocTimes, linkFunction):
        super().__init__(svEmbeddingAllTimes=svEmbeddingAllTimes,
                         svEmbeddingAssocTimes=svEmbeddingAssocTimes,
                        linkFunction=linkFunction)

    def setELLCalculationParams(self, eLLCalculationParams):
        super().setELLCalculationParams(eLLCalculationParams=eLLCalculationParams)
        self._hermQuadWeights = eLLCalculationParams["hermQuadWeights"]
        self._hermQuadPoints = eLLCalculationParams["hermQuadPoints"]

    def _getELinkValues(self, eMean, eVar):
        # aux2 \in  nTrials x nQuadLeg x nNeurons
        aux2 = torch.sqrt(2*eVar)
        # aux3 \in nTrials x nQuadLeg x nNeurons x nQuadLeg
        aux3 = torch.einsum('ijk,l->ijkl', aux2, torch.squeeze(self._hermQuadPoints))
        # aux4 \in nQuad x nQuadLeg x nTrials x nQuadLeg
        aux4 = torch.add(input=aux3, other=eMean.unsqueeze(dim=3))
        # aux5 \in nQuad x nQuadLeg x nTrials x nQuadLeg
        aux5 = self._linkFunction(input=aux4)
        # intval \in  nTrials x nQuadHerm x nNeurons
        eLinkValues = torch.einsum('ijkl,l->ijk', aux5, self._hermQuadWeights.squeeze())
        return eLinkValues

    def _getELogLinkValues(self, eMean, eVar):
        # log_link = cellvec(cellfun(@(x,y) log(m.link(x + sqrt(2*y).* m.xxHerm'))*m.wwHerm,mu_h_Spikes,var_h_Spikes,'uni',0));
        # aux1[trial] \in nSpikes[trial]
        aux1 = [2*eVar[trial] for trial in range(len(eVar))]
        # aux2[trial] \in nSpikes[trial] x nQuadLeg
        aux2 = [torch.einsum('i,j->ij', aux1[trial].squeeze(), self._hermQuadPoints.squeeze()) for trial in range(len(aux1))]
        # aux3[trial] \in nSpikes[trial] x nQuadLeg
        aux3 = [torch.add(input=aux2[trial],
                          other=torch.unsqueeze(input=eMean[trial], dim=1))
                for trial in range(len(aux2))]
        # aux4[trial] \in nSpikes[trial] x nQuadLeg
        aux4 = [torch.log(input=self._linkFunction(x=aux3[trial])) for trial in range(len(aux3))]
        # aux5[trial] \in nSpikes[trial] x 1
        aux5 = [torch.tensordot(a=aux4[trial], b=self._hermQuadWeights, dims=([1], [0])) for trial in range(len(aux4))]
        eLogLinkValues = torch.cat(tensors=aux5)
        return eLogLinkValues


class PoissonELL(ExpectedLogLikelihood):

    def setELLCalculationParams(self, eLLCalculationParams):
        times = eLLCalculationParams["binTimes"]
        self._binWidth = times[0,1,0]-times[0,0,0]
        self._svEmbeddingAllTimes.setTimes(times=times)

    def computeSVPosteriorOnLatentsStats(self):
        answer = self._svEmbeddingAllTimes.computeSVPosteriorOnLatentsStats()
        return answer

    def evalSumAcrossTrialsAndNeurons(self, svPosteriorOnLatentsStats=None):
        eMean, eVar= self._svEmbeddingAllTimes.\
            computeMeansAndVars(svPosteriorOnLatentsStats=svPosteriorOnLatentsStats)
        eLinkValues, eLogLinkValues = \
                self._getELinkAndELogLinkValues(eMean=eMean, eVar=eVar)
        sELLTerm1 = self._binWidth*eLinkValues.sum()
        # sELLTerm2 = (self._measurements*eLogLinkValues.permute(0, 2, 1)).sum()
        sELLTerm2 = (self._measurements*eLogLinkValues).sum()
        answer = -sELLTerm1+sELLTerm2
        return answer

    def buildKernelsMatrices(self):
        self._svEmbeddingAllTimes.buildKernelsMatrices()

    def computeSVPostOnLatentsStats(self):
        answer = self._svEmbeddingAllTimes.computeSVPostOnLatentsStats()
        return answer

    def setMeasurements(self, measurements):
        # measurements \in nTrials x nNeurons x maxNBins
        self._measurements = measurements

    def setIndPointsLocs(self, locs):
        self._svEmbeddingAllTimes.setIndPointsLocs(locs=locs)
        self._svEmbeddingAssocTimes.setIndPointsLocs(locs=locs)

    def setKernels(self, kernels):
        self._svEmbeddingAllTimes.\
            setKernels(kernels=kernels)

    def setInitialParams(self, initial_params):
        self._svEmbeddingAllTimes.\
            setInitialParams(initial_params=initial_params)

    @abstractmethod
    def _getELinkAndELogLinkValues(self, eMean, eVar):
        pass

class PoissonELLExpLink(PoissonELL):

    def __init__(self, svEmbeddingAllTimes):
        super().__init__(svEmbeddingAllTimes=svEmbeddingAllTimes,
                         linkFunction=torch.exp)

    def _getELinkAndELogLinkValues(self, eMean, eVar):
        # intval \in nTrials x nBins x nNeurons
        eLinkValues = self._linkFunction(input=eMean+0.5*eVar)
        eLogLinkValues = eMean
        return eLinkValues, eLogLinkValues

class PoissonELLQuad(PoissonELL):

    def __init__(self, svEmbeddingAllTimes, linkFunction):
        super().__init__(svEmbeddingAllTimes=svEmbeddingAllTimes,
                         linkFunction=linkFunction)

    def _getELinkAndELogLinkValues(self, eMean, eVar):
        # intval = permute(mtimesx(m.wwHerm',permute(m.link(qHmeanAtQuad + sqrt(2*qHMVarAtQuad).*permute(m.xxHerm,[2 3 4 1])),[4 1 2 3])),[2 3 4 1]);

        # aux2 \in  nTrials x maxNBins x nNeurons
        aux2 = torch.sqrt(2*eVar)
        # aux3 \in nTrials x maxNBins x nNeurons x nQuadLeg
        aux3 = torch.einsum('ijk,l->ijkl', aux2, torch.squeeze(self._hermQuadPoints))
        # aux4 \in maxNTrials x maxNBins x nNeurons x nQuadLeg
        aux4 = torch.add(input=aux3, other=eMean.unsqueeze(dim=3))
        # aux5a \in maxNTrials x maxNBins x nNeurons x nQuadLeg
        aux5a = self._linkFunction(input=aux4)
        aux5b = aux5a.log()
        # intval \in  maxNTrials x maxNBins x nQuadLeg
        eLinkValues = torch.einsum('ijkl,l->ijk', aux5b, self._hermQuadWeights.squeeze())
        eLoglinkValues = torch.einsum('ijkl,l->ijk', aux5a, self._hermQuadWeights.squeeze())
        return eLinkValues, eLogLinkValues





from abc import ABC, abstractmethod
import torch
import scipy.stats
import svGPFA.stats.kernelsMatricesStore


class SVPosteriorOnLatents(ABC):

    def __init__(self, svPosteriorOnIndPoints, indPointsLocsKMS,
                 indPointsLocsAndTimesKMS):
        self._svPosteriorOnIndPoints = svPosteriorOnIndPoints
        self._indPointsLocsKMS = indPointsLocsKMS
        self._indPointsLocsAndTimesKMS = indPointsLocsAndTimesKMS

    @abstractmethod
    def computeMeansAndVars(self):
        pass

    @abstractmethod
    def buildKernelsMatrices(self):
        pass

    @abstractmethod
    def setKernels(self):
        pass

    @abstractmethod
    def setInitialParams(self, initial_params):
        pass

    def setTimes(self, times):
        self._indPointsLocsAndTimesKMS.setTimes(times=times)

    @abstractmethod
    def setIndPointsLocs(self, indPointsLocs):
        pass

    def getSVPosteriorOnIndPointsParams(self):
        return self._svPosteriorOnIndPoints.getParams()

    def getKernels(self):
        return self._indPointsLocsKMS.getKernels()

    def getKernelsParams(self):
        return self._indPointsLocsKMS.getKernelsParams()

    def getIndPointsLocs(self):
        return self._indPointsLocsKMS.getIndPointsLocs()

    def setPriorCovRegParam(self, priorCovRegParam):
        self._indPointsLocsKMS.setRegParam(reg_param=priorCovRegParam)


class SVPosteriorOnLatentsAllTimes(SVPosteriorOnLatents):

    def predict(self, times):
        # times \in n_trials x n_times x 1
        # Note: forcing all trials to have the same number of time samples is a
        # limitation of the current version of the code

        kernels = self._indPointsLocsKMS.getKernels()

        indPointsLocsAndAllTimesKMS = \
            svGPFA.stats.kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
        indPointsLocsAndAllTimesKMS.setKernels(kernels=kernels)
        indPointsLocsAndAllTimesKMS.setTimes(times=times)
        indPointsLocsAndAllTimesKMS.setIndPointsLocs(
            ind_points_locs=self.getIndPointsLocs())
        indPointsLocsAndAllTimesKMS.buildKernelsMatrices()

        svPosteriorOnLatents = SVPosteriorOnLatentsAllTimes(
            svPosteriorOnIndPoints=self._svPosteriorOnIndPoints,
            indPointsLocsKMS=self._indPointsLocsKMS,
            indPointsLocsAndTimesKMS=indPointsLocsAndAllTimesKMS)
        qKMu, qKVar = svPosteriorOnLatents.computeMeansAndVars()
        return qKMu, qKVar

    def computeMeansAndVars(self):
        Kzz = self._indPointsLocsKMS.getKzz()
        Ktz = self._indPointsLocsAndTimesKMS.getKtz()
        KttDiag = self._indPointsLocsAndTimesKMS.getKttDiag()
        answer = self.__computeMeansAndVarsGivenKernelMatrices(Kzz=Kzz,
                                                               Ktz=Ktz,
                                                               KttDiag=KttDiag)
        return answer

#     def computeMeans(self, times):
#         Kzz = self._indPointsLocsKMS.getKzz()
#         KzzInv = self._indPointsLocsKMS.getKzzInv()
#         indPointsLocsAndAllTimesKMS = \
#             svGPFA.stats.kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
#         indPointsLocsAndAllTimesKMS.setKernels(kernels=self._indPointsLocsKMS.getKernels())
#         indPointsLocsAndAllTimesKMS.setIndPointsLocs(indPointsLocs=self.getIndPointsLocs())
#         indPointsLocsAndAllTimesKMS.setTimes(times=times)
#         indPointsLocsAndAllTimesKMS.buildKernelsMatrices()
#         Ktz = indPointsLocsAndAllTimesKMS.getKtz()
#         answer = self.__computeMeansGivenKernelMatrices(Kzz=Kzz,
#                                                         KzzInv=KzzInv,
#                                                         Ktz=Ktz)
#         return answer

    def computeMeansAndVarsAtTimes(self, times):
        Kzz = self._indPointsLocsKMS.getKzz()

        indPointsLocsAndAllTimesKMS = \
            svGPFA.stats.kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
        indPointsLocsAndAllTimesKMS.setKernels(
            kernels=self._indPointsLocsKMS.getKernels())
        indPointsLocsAndAllTimesKMS.setTimes(times=times)
        indPointsLocsAndAllTimesKMS.setIndPointsLocs(
            indPointsLocs=self.getIndPointsLocs())
        indPointsLocsAndAllTimesKMS.buildKernelsMatrices()

        Ktz = indPointsLocsAndAllTimesKMS.getKtz()
        KttDiag = indPointsLocsAndAllTimesKMS.getKttDiag()
        answer = self.__computeMeansAndVarsGivenKernelMatrices(Kzz=Kzz,
                                                               Ktz=Ktz,
                                                               KttDiag=KttDiag)

        return answer

    def sample(self, times, nudget=1e-3):
        Kzz = self._indPointsLocsKMS.getKzz()
        # KzzInv = self._indPointsLocsKMS.getKzzInv()

        indPointsLocsAndAllTimesKMS = \
            svGPFA.stats.kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
        indPointsLocsAndAllTimesKMS.setKernels(
            kernels=self._indPointsLocsKMS.getKernels())
        indPointsLocsAndAllTimesKMS.setIndPointsLocs(
            indPointsLocs=self._indPointsLocsKMS.getIndPointsLocs())
        indPointsLocsAndAllTimesKMS.setTimes(times=times)
        indPointsLocsAndAllTimesKMS.buildKernelsMatrices()
        indPointsLocsAndAllTimesKMS.buildKttKernelsMatrices()
        Ktz = indPointsLocsAndAllTimesKMS.getKtz()
        Ktt = indPointsLocsAndAllTimesKMS.getKtt()

        qMu = self._svPosteriorOnIndPoints.getMean()
        qSigma = self._svPosteriorOnIndPoints.buildCov()

        nLatents = len(Kzz)
        nTrials = Kzz[0].shape[0]
        samples = [[] for r in range(nTrials)]
        means = [[] for r in range(nTrials)]
        variances = [[] for r in range(nTrials)]
        for r in range(nTrials):
            samples[r] = torch.empty((nLatents, Ktt[0].shape[1]),
                                     dtype=Kzz[0].dtype)
            means[r] = torch.empty((nLatents, Ktt[0].shape[1]),
                                   dtype=Kzz[0].dtype)
            variances[r] = torch.empty((nLatents, Ktt[0].shape[1]),
                                       dtype=Kzz[0].dtype)
            for k in range(nLatents):
                print("Processing trial {:d} and latent {:d}".format(r, k))
                Kzzrk = Kzz[k][r, :, :]
                # KzzInvrk = KzzInv[k][r,:,:]
                Ktzrk = Ktz[k][r, :, :]
                Kttrk = Ktt[k][r, :, :]
                qMurk = qMu[k][r, :, :]
                qSigmark = qSigma[k][r, :, :]

                # begin compute mean #
                # b = torch.cholesky_solve(qMurk, KzzInvrk)
                b = self._indPointsLocsKMS.solveForLatentAndTrial(
                    input=qMurk, latentIndex=k, trialIndex=r)
                meanrk = torch.squeeze(Ktzrk.matmul(b)).detach().numpy()
                # end compute mean #

                # being compute covar #
                # B = torch.cholesky_solve(torch.t(Ktzrk), KzzInvrk)
                B = self._indPointsLocsKMS.solveForLatentAndTrial(
                    input=torch.t(Ktzrk), latentIndex=k, trialIndex=r)
                covarrk = Kttrk+torch.t(B).matmul(qSigmark-Kzzrk).matmul(B)
                # end compute covar #

                covarrk += torch.eye(covarrk.shape[0])*nudget
                covarrk = covarrk.detach().numpy()
                mn = scipy.stats.multivariate_normal(mean=meanrk, cov=covarrk)
                samples[r][k, :] = torch.from_numpy(mn.rvs(size=1))
                means[r][k, :] = torch.from_numpy(meanrk)
                variances[r][k, :] = torch.diag(torch.from_numpy(covarrk))
        return samples, means, variances

    def __computeMeansAndVarsGivenKernelMatrices(self, Kzz, Ktz, KttDiag):
        nTrials = KttDiag.shape[0]
        nQuad = KttDiag.shape[1]
        nLatent = KttDiag.shape[2]

        qKMu = torch.empty((nTrials, nQuad, nLatent), dtype=Kzz[0].dtype,
                           device=Kzz[0].device)
        qKVar = torch.empty((nTrials, nQuad, nLatent), dtype=Kzz[0].dtype,
                            device=Kzz[0].device)

        qSigma = self._svPosteriorOnIndPoints.buildCov()
        for k in range(len(self._svPosteriorOnIndPoints.getMean())):
            # Ak \in nTrials x nInd[k] x 1
            Ak = self._indPointsLocsKMS.solveForLatent(
                input=self._svPosteriorOnIndPoints.getMean()[k], latentIndex=k)
            # qKMu \in  nTrial x nQuad x nLatent
            qKMu[:, :, k] = torch.squeeze(torch.matmul(Ktz[k], Ak))

            # Bkf \in nTrials x nInd[k] x nQuad
            Bkf = self._indPointsLocsKMS.solveForLatent(
                input=Ktz[k].transpose(dim0=1, dim1=2), latentIndex=k)
            # mm1f \in nTrials x nInd[k] x nQuad
            mm1f = torch.matmul(qSigma[k]-Kzz[k], Bkf)
            # aux1 \in nTrials x nInd[k] x nQuad
            aux1 = Bkf*mm1f
            # aux2 \in nTrials x nQuad
            aux2 = torch.sum(input=aux1, dim=1)
            # aux3 \in nTrials x nQuad
            aux3 = KttDiag[:, :, k]+aux2
            # qKVar \in nTrials x nQuad x nLatent
            qKVar[:, :, k] = aux3
        return qKMu, qKVar

    def __computeMeansGivenKernelMatrices(self, Kzz, KzzInv, Ktz):
        nTrials = Ktz[0].shape[0]
        nQuad = Ktz[0].shape[1]
        nLatents = len(Ktz)

        qKMu = torch.empty((nTrials, nQuad, nLatents), dtype=Kzz[0].dtype,
                           device=Kzz[0].device)
        for k in range(nLatents):
            # Ak \in nTrials x nInd[k] x 1
            Ak = self._indPointsLocsKMS.solveForLatent(
                input=self._svPosteriorOnIndPoints.getMean()[k], latentIndex=k)
            # qKMu \in  nTrial x nQuad x nLatents
            qKMu[:, :, k] = torch.squeeze(torch.matmul(Ktz[k], Ak))
        return qKMu

    def buildKernelsMatrices(self):
        self._indPointsLocsKMS.buildKernelsMatrices()
        self._indPointsLocsAndTimesKMS.buildKernelsMatrices()

    def setKernels(self, kernels):
        self._indPointsLocsKMS.setKernels(kernels=kernels)
        self._indPointsLocsAndTimesKMS.setKernels(kernels=kernels)

    def setInitialParams(self, initial_params):
        self._svPosteriorOnIndPoints.setInitialParams(
            initial_params=initial_params["posterior_on_ind_points"])
        self._indPointsLocsKMS.setInitialParams(
            initial_params=initial_params["kernels_matrices_store"])
        self._indPointsLocsAndTimesKMS.setInitialParams(
            initial_params=initial_params["kernels_matrices_store"])

    def setIndPointsLocs(self, indPointsLocs):
        self._indPointsLocsKMS.setIndPointsLocs(ind_pointsLocs=indPointsLocs)
        self._indPointsLocsAndTimesKMS.setIndPointsLocs(
            ind_points_locs=indPointsLocs)


class SVPosteriorOnLatentsAssocTimes(SVPosteriorOnLatents):

    def computeMeansAndVars(self):
        Kzz = self._indPointsLocsKMS.getKzz()
        Ktz = self._indPointsLocsAndTimesKMS.getKtz()
        KttDiag = self._indPointsLocsAndTimesKMS.getKttDiag()
        qKMu, qKVar = self.__computeMeansAndVarsGivenKernelMatrices(
            Kzz=Kzz, Ktz=Ktz, KttDiag=KttDiag)
        return qKMu, qKVar

    def __computeMeansAndVarsGivenKernelMatrices(self, Kzz, Ktz, KttDiag):
        nTrials = len(KttDiag[0])
        nLatent = len(self._svPosteriorOnIndPoints.getMean())
        # Ak[k] \in nTrial x nInd[k] x 1
        Ak = [self._indPointsLocsKMS.solveForLatent(
            input=self._svPosteriorOnIndPoints.getMean()[k], latentIndex=k)
            for k in range(nLatent)]
        qSigma = self._svPosteriorOnIndPoints.buildCov()
        qKMu = [[None] for tr in range(nTrials)]
        qKVar = [[None] for tr in range(nTrials)]
        for trialIndex in range(nTrials):
            nSpikesForTrial = KttDiag[0][trialIndex].shape[0]
            # qKMu[trialIndex] \in nSpikesForTrial[trialIndex] x nLatent
            qKMu[trialIndex] = torch.empty((nSpikesForTrial, nLatent),
                                           dtype=Kzz[0].dtype,
                                           device=Kzz[0].device)
            qKVar[trialIndex] = torch.empty((nSpikesForTrial, nLatent),
                                            dtype=Kzz[0].dtype,
                                            device=Kzz[0].device)
            for k in range(nLatent):
                qKMu[trialIndex][:, k] = torch.squeeze(
                    torch.mm(input=Ktz[k][trialIndex],
                             mat2=Ak[k][trialIndex, :, :]))
                # Bfk \in nInd[k] x nSpikesForTrial[trialIndex]
                Bfk = self._indPointsLocsKMS.solveForLatentAndTrial(
                    input=Ktz[k][trialIndex].transpose(dim0=0, dim1=1),
                    latentIndex=k, trialIndex=trialIndex)
                # mm1f \in nInd[k] x nSpikesForTrial[trialIndex]
                diff = qSigma[k][trialIndex, :, :]-Kzz[k][trialIndex, :, :]
                mm1f = torch.matmul(diff, Bfk)
                # qKVar[trialIndex] \in nSpikesForTrial[trialIndex] x nLatent
                qKVar[trialIndex][:, k] = \
                    torch.squeeze(KttDiag[k][trialIndex]) + \
                    torch.sum(a=Bfk*mm1f, axis=0)

        return qKMu, qKVar

    def buildKernelsMatrices(self):
        # not asking _indPointsLocsKMS to buildMatrices because
        # SVPosteriorOnLatentsAlTimes already did so
        self._indPointsLocsAndTimesKMS.buildKernelsMatrices()

    def setKernels(self, kernels):
        self._indPointsLocsAndTimesKMS.setKernels(kernels=kernels)

    def setInitialParams(self, initial_params):
        self._indPointsLocsAndTimesKMS.setInitialParams(
            initial_params=initial_params["kernels_matrices_store"])

    def setIndPointsLocs(self, indPointsLocs):
        # not asking _indPointsLocsKMS to setIndPointsLocs becasue
        # SVPosteriorOnLatentsAlTimes already did so
        self._indPointsLocsAndTimesKMS.setIndPointsLocs(
            ind_points_locs=indPointsLocs)




import warnings
import torch
import scipy.stats
import gcnu_common.stats.gaussianProcesses.eval
import gcnu_common.stats.pointProcesses.sampling
from ..stats import kernelsMatricesStore


class BaseSimulator:

    def getCIF(self, n_trials, latents_samples, C, d, link_function):
        n_neurons = C.shape[0]
        cif_values = [[] for n in range(n_trials)]
        for r in range(n_trials):
            embeddings = torch.matmul(C, latents_samples[r]) + d
            cif_values[r] = [link_function(embeddings[n, :])
                             for n in range(n_neurons)]
        return(cif_values)

    def simulate(self, cif_trials_times, cif_values,
                 sampling_func=gcnu_common.stats.pointProcesses.sampling.sampleInhomogeneousPP_thinning):
        n_trials = len(cif_values)
        n_neurons = len(cif_values[0])
        spikes_times = [[] for n in range(n_trials)]
        for r in range(n_trials):
            spikes_times[r] = [[] for r in range(n_neurons)]
            for n in range(n_neurons):
                print("Processing trial {:d} and neuron {:d}".format(r, n))
                spikes_times[r][n] = torch.tensor(
                    sampling_func(CIF_times=cif_trials_times[r],
                                  CIF_values=cif_values[r][n])["inhomogeneous"])
        return(spikes_times)


class GPFASimulator(BaseSimulator):
    def getLatentsSamplesMeansAndSTDs(self, n_trials, means_funcs, kernels,
                                      trials_times, prior_cov_reg_param,
                                      dtype):
        n_latents = len(kernels)
        latents_samples = [[] for r in range(n_trials)]
        latents_means = [[] for r in range(n_trials)]
        latents_STDs = [[] for r in range(n_trials)]

        for r in range(n_trials):
            latents_samples[r] = torch.empty((n_latents, len(trials_times[r])),
                                             dtype=dtype)
            latents_means[r] = torch.empty((n_latents, len(trials_times[r])),
                                           dtype=dtype)
            latents_STDs[r] = torch.empty((n_latents, len(trials_times[r])),
                                          dtype=dtype)
            for k in range(n_latents):
                print("Processing trial {:d} and latent {:d}".format(r, k))
                gp = gcnu_common.stats.gaussianProcesses.eval.GaussianProcess(
                    mean=means_funcs[k], kernel=kernels[k])
                sample, mean, cov = gp.eval(
                    t=trials_times[r], regularization=prior_cov_reg_param)
                latents_samples[r][k, :] = sample
                latents_means[r][k, :] = mean
                latents_STDs[r][k, :] = torch.diag(cov).sqrt()
        return latents_samples, latents_means, latents_STDs


class GPFAwithIndPointsSimulator(BaseSimulator):

    def getLatentsSamplesMeansAndSTDs(self, var_mean, var_cov, kernels,
                                      ind_points_locs, trials_times,
                                      prior_cov_reg_param,
                                      latents_cov_reg_param, dtype):
        n_latents = len(kernels)
        n_trials = ind_points_locs[0].shape[0]

        indPointsLocsKMS = kernelsMatricesStore.IndPointsLocsKMS_Chol()
        indPointsLocsKMS.setKernels(kernels=kernels)
        indPointsLocsKMS.setIndPointsLocs(ind_points_locs=ind_points_locs)
        indPointsLocsKMS.setRegParam(reg_param=prior_cov_reg_param)
        indPointsLocsKMS.buildKernelsMatrices()
        Kzz = indPointsLocsKMS.getKzz()
        condNumberThr = 1e+6
        eigRes = torch.eig(Kzz[0][0, :, :], eigenvectors=True)
        if any(eigRes.eigenvalues[:, 1] > 0):
            warnings.warn("Some eigenvalues of Kzz are imaginary")
        sortedEigVals = eigRes.eigenvalues[:, 0].sort(descending=True).values
        condNumber = sortedEigVals[0]/sortedEigVals[-1]
        if condNumber > condNumberThr:
            warnings.warn("Poorly conditioned Kzz "
                          f"(condition number={condNumber}")

        indPointsLocsAndAllTimesKMS = \
            kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
        indPointsLocsAndAllTimesKMS.setKernels(kernels=kernels)
        indPointsLocsAndAllTimesKMS.setIndPointsLocs(
            ind_points_locs=ind_points_locs)
        indPointsLocsAndAllTimesKMS.setTimes(times=trials_times)
        indPointsLocsAndAllTimesKMS.buildKernelsMatrices()
        Ktz = indPointsLocsAndAllTimesKMS.getKtz()

        latents_samples = [[] for r in range(n_trials)]
        latents_means = [[] for r in range(n_trials)]
        latents_STDs = [[] for r in range(n_trials)]
        for r in range(n_trials):
            latents_samples[r] = torch.empty((n_latents,
                                              trials_times.shape[1]),
                                             dtype=torch.double)
            latents_means[r] = torch.empty((n_latents, trials_times.shape[1]),
                                           dtype=torch.double)
            latents_STDs[r] = torch.empty((n_latents, trials_times.shape[1]),
                                          dtype=torch.double)
            for k in range(n_latents):
                # first let's compute the variational mean
                tmp = indPointsLocsKMS.solveForLatentAndTrial(
                    input=var_mean[r][k], latentIndex=k, trialIndex=r)
                mean = torch.matmul(Ktz[k][r, :, :], tmp).squeeze()
                # done with the variational mean

                # not let's compute the variational covariance
                # cov = Ktt + Ktz Kzz^{-1} (Sk-Kzz) Kzz^{-1} Kst
                # cov = Ktt + Ktz C
                # A = solve(Kzz, Kzt)
                # B = (Sk-Kzz) A
                # C = solze(Kzz, B)
                A = indPointsLocsKMS.solveForLatentAndTrial(
                    input=Ktz[k][r, :, :].T, latentIndex=k, trialIndex=r)
                B = torch.matmul(var_cov[r][k]-Kzz[k][r, :, :], A)
                C = indPointsLocsKMS.solveForLatentAndTrial(
                    input=B, latentIndex=k, trialIndex=r)
                cov = kernels[k].buildKernelMatrix(trials_times[r, :, 0]) + \
                    torch.matmul(Ktz[k][r, :, :], C)
                # done with the variational cov

                std = torch.diag(cov).sqrt()
                cov += torch.eye(cov.shape[0])*latents_cov_reg_param
                mn = scipy.stats.multivariate_normal(mean=mean, cov=cov)
                sample = torch.from_numpy(mn.rvs())
                latents_samples[r][k, :] = sample
                latents_means[r][k, :] = mean
                latents_STDs[r][k, :] = std
        return latents_samples, latents_means, latents_STDs, Kzz




import pdb
import torch
import warnings

class SVLowerBound:

    def __init__(self, eLL, klDiv):
        super(SVLowerBound, self).__init__()
        self._eLL = eLL
        self._klDiv = klDiv

    def setParamsAndData(self, measurements, initial_params,
                         eLLCalculationParams, priorCovRegParam):
        """Sets model parameters and data.

        :param measurements: ``measurements[r][n]`` are the measurements for trial ``r`` and neuron ``n``.

            For a point-process SVLowerBound (i.e., for a SVLowerBound constructed with an expected log-likelihood of class :class:`svGPFA.stats.expectedLogLikelihood.PointProcessELL`) ``measurements[r][n]`` should be a list of spikes times for trial ``r`` and neuron ``n``.

            For a Poisson SVLowerBound (i.e., for a SVLowerBound constructed with an expected log-likelihood of class :class:`svGPFA.stats.expectedLogLikelihood.PoissonELL`) ``measurements[r][n]`` should be a tuple of length ``n_bins`` containing the spike counts in bins.

        :type  measurements: nested list

        :param initial_params: initial parameters as returned by :func:`svGPFA.utils.initUtils.getParamsAndKernelsTypes`.

        :type  initial_params: dictionary

        :param eLLCalculationParams: parameters used to calculate the expected log likelighood.

            For a point-process SVLowerBound ``eLLCalculationParams`` should be a dictionary with keys ``leg_quad_points`` and ``leg_quad_points``, containing the Legendre quadrature points and weights, respectivey, used to calculate the integral of the expected log likelihood (Eq. 7 in :cite:t:`dunckerAndSahani18`).

            For a Poisson SVLowerBound ``eLLCalculationParams`` should be a dictionary with key ``binTimes`` containing the mean time of every bin.

        :type eLLCalculationParams: dictionary

        :param priorCovRegParam: regularization parameter for the prior covariance matrix (:math:`K_{zz}` in Eq. 2 of :cite:t:`dunckerAndSahani18`).

        :type priorCovRegParam: float
        """
        self.setMeasurements(measurements=measurements)
        self.setInitialParams(initial_params=initial_params)
        self.setELLCalculationParams(eLLCalculationParams=eLLCalculationParams)
        self.setPriorCovRegParam(priorCovRegParam=priorCovRegParam)
        self.buildKernelsMatrices()

    def eval(self):
        eLLEval = self._eLL.evalSumAcrossTrialsAndNeurons()
        klDivEval = self._klDiv.evalSumAcrossLatentsAndTrials()
        theEval = eLLEval-klDivEval
        if torch.isinf(theEval):
            # raise RuntimeError("infinity lower bound detected")
            warnings.warn("infinity lower bound detected")
        return theEval

    def sampleCIFs(self, times, nudget=1e-3):
        answer = self._eLL.sampleCIFs(times=times, nudget=nudget)
        return answer

#     def computeCIFsMeans(self, times):
#         answer = self._eLL.computeCIFsMeans(times=times)
#         return answer
# 
    def computeExpectedPosteriorCIFs(self, times):
        answer = self._eLL.computeExpectedPosteriorCIFs(times=times)
        return answer

    def computeEmbeddingMeansAndVarsAtTimes(self, times):
        answer = self._eLL.computeEmbeddingsMeansAndVarsAtTimes(times=times)
        return answer

    def evalELLSumAcrossTrialsAndNeurons(self, svPosteriorOnLatentsStats):
        answer = self._eLL.evalSumAcrossTrialsAndNeurons(
            svPosteriorOnLatentsStats=svPosteriorOnLatentsStats)
        return answer

    def buildKernelsMatrices(self):
        self._eLL.buildKernelsMatrices()

    def computeSVPosteriorOnLatentsStats(self):
        return self._eLL.computeSVPosteriorOnLatentsStats()

    def setInitialParams(self, initial_params):
        self._eLL.setInitialParams(initial_params=initial_params)

    def setKernels(self, kernels):
        self._eLL.setKernels(kernels=kernels)

    def setMeasurements(self, measurements):
        self._eLL.setMeasurements(measurements=measurements)

    def setIndPointsLocs(self, locs):
        self._eLL.setIndPointsLocs(locs=locs)

    def setPriorCovRegParam(self, priorCovRegParam):
        self._eLL.setPriorCovRegParam(priorCovRegParam=priorCovRegParam)

    def setELLCalculationParams(self, eLLCalculationParams):
        self._eLL.setELLCalculationParams(eLLCalculationParams=eLLCalculationParams)

    def getSVPosteriorOnIndPointsParams(self):
        return self._eLL.getSVPosteriorOnIndPointsParams()

    def getSVEmbeddingParams(self):
        return self._eLL.getSVEmbeddingParams()

    def getKernels(self):
        return self._eLL.getKernels()

    def getKernelsParams(self):
        return self._eLL.getKernelsParams()

    def getIndPointsLocs(self):
        return self._eLL.getIndPointsLocs()

    def predictLatents(self, times):
        return self._eLL.predictLatents(times=times)

    def predictEmbedding(self, times):
        return self._eLL.predictEmbedding(times=times)


class SVLowerBoundWithParamsGettersAndSetters(SVLowerBound):

    def __init__(self, eLL, klDiv):
        super(SVLowerBoundWithParamsGettersAndSetters, self).__init__(
            eLL, klDiv)
        # shortcuts
        self._svPosteriorOnIndPoints = klDiv.get_svPosteriorOnIndPoints()
        self._svEmbedding = eLL.get_svEmbeddingAllTimes()
        self._indPointsLocsKMS = klDiv.get_indPointsLocsKMS()

    # svPosterioOnIndPoints
    def get_flattened_svPosteriorOnIndPoints_params(self):
        flattened_params = self._svPosteriorOnIndPoints.get_flattened_params()
        return flattened_params

    def get_flattened_svPosteriorOnIndPoints_params_grad(self):
        flattened_params_grad = self._svPosteriorOnIndPoints.get_flattened_params_grad()
        return flattened_params_grad

    def set_svPosteriorOnIndPoints_params_from_flattened(self, flattened_params):
        self._svPosteriorOnIndPoints.set_params_from_flattened(flattened_params=flattened_params)

    def set_svPosteriorOnIndPoints_params_requires_grad(self, requires_grad):
        self._svPosteriorOnIndPoints.set_params_requires_grad(requires_grad=requires_grad)

    # svEmbedding
    def get_flattened_svEmbedding_params(self):
        flattened_params = self._svEmbedding.get_flattened_params()
        return flattened_params

    def get_flattened_svEmbedding_params_grad(self):
        flattened_params_grad = self._svEmbedding.get_flattened_params_grad()
        return flattened_params_grad

    def set_svEmbedding_params_from_flattened(self, flattened_params):
        self._svEmbedding.set_params_from_flattened(flattened_params=flattened_params)

    def set_svEmbedding_params_requires_grad(self, requires_grad):
        self._svEmbedding.set_params_requires_grad(requires_grad=requires_grad)

    # kernels_params
    def get_flattened_kernels_params(self):
        flattened_params = self._indPointsLocsKMS.get_flattened_kernels_params()
        return flattened_params

    def get_flattened_kernels_params_grad(self):
        flattened_params_grad = self._indPointsLocsKMS.get_flattened_kernels_params_grad()
        return flattened_params_grad

    def set_kernels_params_from_flattened(self, flattened_params):
        self._indPointsLocsKMS.set_kernels_params_from_flattened(flattened_params=flattened_params)

    def set_kernels_params_requires_grad(self, requires_grad):
        self._indPointsLocsKMS.set_kernels_params_requires_grad(requires_grad=requires_grad)

    # indPointsLocs
    def get_flattened_indPointsLocs(self):
        flattened_params = self._indPointsLocsKMS.get_flattened_indPointsLocs()
        return flattened_params

    def get_flattened_indPointsLocs_grad(self):
        flattened_params_grad = self._indPointsLocsKMS.get_flattened_indPointsLocs_grad()
        return flattened_params_grad

    def set_indPointsLocs_from_flattened(self, flattened_params):
        self._indPointsLocsKMS.set_indPointsLocs_from_flattened(flattened_params=flattened_params)

    def set_indPointsLocs_requires_grad(self, requires_grad):
        self._indPointsLocsKMS.set_indPointsLocs_requires_grad(requires_grad=requires_grad)





import pdb
from abc import ABC, abstractmethod
import math
import torch

import svGPFA.utils.miscUtils

class SVPosteriorOnIndPoints(ABC):

    @abstractmethod
    def setInitialParams(self, initial_params):
        pass

    @abstractmethod
    def getParams(self):
        pass

    @abstractmethod
    def getMean(self):
        pass

    @abstractmethod
    def buildCov(self):
        pass

class SVPosteriorOnIndPointsChol(SVPosteriorOnIndPoints):

    def __init__(self):
        super(SVPosteriorOnIndPoints, self).__init__()

    def setInitialParams(self, initial_params):
        nLatents = len(initial_params["mean"])
        self._mean = [initial_params["mean"][k] for k in range(nLatents)]
        self._cholVecs = [initial_params["cholVecs"][k] for k in range(nLatents)]

    def getParams(self):
        listOfTensors = []
        listOfTensors.extend([self._mean[k] for k in range(len(self._mean))])
        listOfTensors.extend([self._cholVecs[k] for k in
                              range(len(self._cholVecs))])
        return listOfTensors

    def getMean(self):
        return self._mean

    def buildCov(self):
        qSigma = svGPFA.utils.miscUtils.buildCovsFromCholVecs(cholVecs=self._cholVecs)
        return qSigma

class SVPosteriorOnIndPointsCholWithGettersAndSetters(SVPosteriorOnIndPointsChol):
    def get_flattened_params(self):
        flattened_params = []
        for k in range(len(self._mean)):
            flattened_params.extend(self._mean[k].flatten().tolist())
        for k in range(len(self._cholVecs)):
            flattened_params.extend(self._cholVecs[k].flatten().tolist())
        return flattened_params

    def get_flattened_params_grad(self):
        flattened_params_grad = []
        for k in range(len(self._mean)):
            flattened_params_grad.extend(self._mean[k].grad.flatten().tolist())
        for k in range(len(self._cholVecs)):
            flattened_params_grad.extend(self._cholVecs[k].grad.flatten().tolist())
        return flattened_params_grad

    def set_params_from_flattened(self, flattened_params):
        for k in range(len(self._mean)):
            flattened_param = flattened_params[:self._mean[k].numel()]
            self._mean[k] = torch.tensor(flattened_param, dtype=torch.double).reshape(self._mean[k].shape)
            flattened_params = flattened_params[self._mean[k].numel():]
        for k in range(len(self._cholVecs)):
            flattened_param = flattened_params[:self._cholVecs[k].numel()]
            self._cholVecs[k] = torch.tensor(flattened_param,
                                                   dtype=torch.double).reshape(self._cholVecs[k].shape)
            flattened_params = flattened_params[self._cholVecs[k].numel():]

    def set_params_requires_grad(self, requires_grad):
        for k in range(len(self._mean)):
            self._mean[k].requires_grad = requires_grad
        for k in range(len(self._cholVecs)):
            self._cholVecs[k].requires_grad = requires_grad


class SVPosteriorOnIndPointsRank1PlusDiag(SVPosteriorOnIndPoints):

    def __init__(self):
        super(SVPosteriorOnIndPoints, self).__init__()

    def setInitialParams(self, initial_params):
        self._mean = initial_params["mean"]
        self._qSVec = initial_params["qSVec0"]
        self._qSDiag = initial_params["qSDiag0"]

    def getParams(self):
        listOfTensors = []
        listOfTensors.extend([self._mean[k] for k in range(len(self._mean))])
        listOfTensors.extend([self._qSVec[k] for k in range(len(self._qSVec))])
        listOfTensors.extend([self._qSDiag[k]
                              for k in range(len(self._qSDiag))])
        return listOfTensors

    def getMean(self):
        return self._mean

    def buildCov(self):
        R = self._qSVec[0].shape[0]
        K = len(self._qSVec)
        qSigma = [[None] for k in range(K)]
        for k in range(K):
            nIndK = self._qSDiag[k].shape[1]
            # qq \in nTrials x nInd[k] x 1
            qq = self._qSVec[k].reshape(shape=(R, nIndK, 1))
            # dd \in nTrials x nInd[k] x 1
            nIndKVarRnkK = self._qSVec[k].shape[1]
            dd = svGPFA.utils.miscUtils.build3DdiagFromDiagVector(v=(self._qSDiag[k].flatten())**2, M=R, N=nIndKVarRnkK)
            qSigma[k] = torch.matmul(qq, torch.transpose(a=qq, dim0=1, dim1=2)) + dd
        return(qSigma)





import pdb
import abc
import torch

class SVEmbedding(abc.ABC):

    def __init__(self, svPosteriorOnLatents):
        self._svPosteriorOnLatents = svPosteriorOnLatents

    def computeMeansAndVars(self, svPosteriorOnLatentsStats=None):
        if svPosteriorOnLatentsStats is None:
            svPosteriorOnLatentsStats = \
                self._svPosteriorOnLatents.computeMeansAndVars()
        means, vars = self._computeMeansAndVarsGivenSVPosteriorOnLatentsStats(
            means=svPosteriorOnLatentsStats[0],
            vars=svPosteriorOnLatentsStats[1])
        return means, vars

    def computeSVPosteriorOnLatentsStats(self):
        return self._svPosteriorOnLatents.computeMeansAndVars()

    def buildKernelsMatrices(self):
        self._svPosteriorOnLatents.buildKernelsMatrices()

    @abc.abstractmethod
    def setInitialParams(self, initial_params):
        pass

    def setIndPointsLocs(self, indPointsLocs):
        self._svPosteriorOnLatents.setIndPointsLocs(indPointsLocs=indPointsLocs)

    def setKernels(self, kernels):
        self._svPosteriorOnLatents.setKernels(kernels=kernels)

    def setTimes(self, times):
        self._svPosteriorOnLatents.setTimes(times=times)

    @abc.abstractmethod
    def getParams(self):
        pass

    def getSVPosteriorOnIndPointsParams(self):
        return self._svPosteriorOnLatents.getSVPosteriorOnIndPointsParams()

    def getIndPointsLocs(self):
        return self._svPosteriorOnLatents.getIndPointsLocs()

    def getKernels(self):
        return self._svPosteriorOnLatents.getKernels()

    def getKernelsParams(self):
        return self._svPosteriorOnLatents.getKernelsParams()

    @abc.abstractmethod
    def _computeMeansAndVarsGivenSVPosteriorOnLatentsStats(self, means, vars):
        pass

class LinearSVEmbedding(SVEmbedding):

    def setInitialParams(self, initial_params):
        svEmbeddingInitialParams = initial_params["embedding"]
        self._C = svEmbeddingInitialParams["C0"]
        self._d = svEmbeddingInitialParams["d0"]
        svPosteriorOnLatentsInitialParams = initial_params["posterior_on_latents"]
        self._svPosteriorOnLatents.setInitialParams(
            initial_params=svPosteriorOnLatentsInitialParams)

    def sample(self, times, nudget=1e-3):
        latentsSamples, _, _ = self._svPosteriorOnLatents.sample(times=times,
                                                                 nudget=nudget)
        answer = [self._C.matmul(latentsSamples[r])+self._d for r in range(len(latentsSamples))]
        return answer

    def getParams(self):
        return [self._C, self._d]

class LinearSVEmbeddingAllTimes(LinearSVEmbedding):

    def _computeMeansAndVarsGivenSVPosteriorOnLatentsStats(self, means, vars):
        # emb_post_mean, emb_post_var \in nTrials x nQuad x nNeurons
        emb_post_mean = torch.matmul(means, torch.t(self._C)) + torch.reshape(input=self._d, shape=(1, 1, len(self._d))) # using broadcasting
        emb_post_var = torch.matmul(vars, (torch.t(self._C))**2)
        return emb_post_mean, emb_post_var

    def predictLatents(self, times):
        return self._svPosteriorOnLatents.predict(times=times)

    def predict(self, times):
        qKMu, qKVar = self._svPosteriorOnLatents.predict(times=times)
        answer = self._computeMeansAndVarsGivenSVPosteriorOnLatentsStats(means=qKMu, vars=qKVar)
        return answer

#     def computeMeans(self, times):
#         qKMu = self._svPosteriorOnLatents.computeMeans(times=times)
#         emb_post_mean = torch.matmul(qKMu, torch.t(self._C)) + torch.reshape(input=self._d, shape=(1, 1, len(self._d))) # using broadcasting
#         return emb_post_mean
# 
    def computeMeansAndVarsAtTimes(self, times):
        qKMu, qKVar = self._svPosteriorOnLatents.computeMeansAndVarsAtTimes(times=times)
        answer = self._computeMeansAndVarsGivenSVPosteriorOnLatentsStats(means=qKMu, vars=qKVar)
        return answer

    def setPriorCovRegParam(self, priorCovRegParam):
        self._svPosteriorOnLatents.setPriorCovRegParam(priorCovRegParam=priorCovRegParam)

class LinearSVEmbeddingAllTimesWithParamsGettersAndSetters(LinearSVEmbeddingAllTimes):
    def get_flattened_params(self):
        flattened_params = []
        flattened_params.extend(self._C.flatten().tolist())
        flattened_params.extend(self._d.flatten().tolist())
        return flattened_params

    def get_flattened_params_grad(self):
        flattened_params_grad = []
        flattened_params_grad.extend(self._C.grad.flatten().tolist())
        flattened_params_grad.extend(self._d.grad.flatten().tolist())
        return flattened_params_grad

    def set_params_from_flattened(self, flattened_params):
        flattened_param = flattened_params[:self._C.numel()]
        self._C = torch.tensor(flattened_param, dtype=torch.double).reshape(self._C.shape)
        flattened_params = flattened_params[self._C.numel():]
        flattened_param = flattened_params[:self._d.numel()]
        self._d = torch.tensor(flattened_param, dtype=torch.double).reshape(self._d.shape)
        flattened_param = flattened_params[self._d.numel():]

    def set_params_requires_grad(self, requires_grad):
        self._C.requires_grad = requires_grad
        self._d.requires_grad = requires_grad

class LinearSVEmbeddingAssocTimes(LinearSVEmbedding):

    def setNeuronForSpikeIndex(self, neuronForSpikeIndex):
        self._neuronForSpikeIndex = neuronForSpikeIndex

    def _computeMeansAndVarsGivenSVPosteriorOnLatentsStats(self, means, vars):
        nTrials = len(self._neuronForSpikeIndex)
        emb_post_mean = [[None] for tr in range(nTrials)]
        emb_post_var = [[None] for tr in range(nTrials)]
        for trialIndex in range(nTrials):
            emb_post_mean[trialIndex] = torch.sum(means[trialIndex]*self._C[(self._neuronForSpikeIndex[trialIndex]).tolist(),:], dim=1)+self._d[(self._neuronForSpikeIndex[trialIndex]).tolist()].squeeze()
            emb_post_var[trialIndex] = torch.sum(vars[trialIndex]*(self._C[(self._neuronForSpikeIndex[trialIndex]).tolist(),:])**2, dim=1)
        return emb_post_mean, emb_post_var





import math
import numpy as np
import torch
import svGPFA.stats.kernels


def getKernels(nLatents, config, forceUnitScale):
    kernels = [[] for r in range(nLatents)]
    for k in range(nLatents):
        kernelType = config["kernel_params"]["kTypeLatent{:d}".format(k)]
        if kernelType=="periodic":
            if not forceUnitScale:
                scale = float(config["kernel_params"]["kScaleValueLatent{:d}".format(k)])
            else:
                scale = 1.0
            lengthscale = float(config["kernel_params"]["kLengthscaleScaledValueLatent{:d}".format(k)])
            period = float(config["kernel_params"]["kPeriodScaledValueLatent{:d}".format(k)])
            kernel = svGPFA.stats.kernels.PeriodicKernel(scale=scale)
            kernel.setParams(params=torch.Tensor([lengthscale, period]).double())
        elif kernelType=="exponentialQuadratic":
            if not forceUnitScale:
                scale = float(config["kernel_params"]["kScaleValueLatent{:d}".format(k)])
            else:
                scale = 1.0
            lengthscale = float(config["kernel_params"]["kLengthscaleScaledValueLatent{:d}".format(k)])
            kernel = svGPFA.stats.kernels.ExponentialQuadraticKernel(scale=scale)
            kernel.setParams(params=torch.Tensor([lengthscale]).double())
        else:
            raise ValueError("Invalid kernel type {:s} for latent {:d}".format(kernelType, k))
        kernels[k] = kernel
    return kernels


def getScaledKernels(nLatents, config, forceUnitScale):
    kernels = [[] for r in range(nLatents)]
    kernelsParamsScales = [[] for r in range(nLatents)]
    for k in range(nLatents):
        kernelType = config["kernel_params"]["kTypeLatent{:d}".format(k)]
        if kernelType == "periodic":
            if not forceUnitScale:
                scale = float(config["kernel_params"]["kScaleValueLatent{:d}".format(k)])
            else:
                scale = 1.0
            lengthscaleScaledValue = float(config["kernel_params"]["kLengthscaleScaledValueLatent{:d}".format(k)])
            lengthscaleScale = float(config["kernel_params"]["kLengthscaleScaleLatent{:d}".format(k)])
            periodScaledValue = float(config["kernel_params"]["kPeriodScaledValueLatent{:d}".format(k)])
            periodScale = float(config["kernel_params"]["kPeriodScaleLatent{:d}".format(k)])
            kernel = svGPFA.stats.kernels.PeriodicKernel(scale=scale, lengthscaleScale=lengthscaleScale, periodScale=periodScale)
            kernel.setParams(params=torch.Tensor([lengthscaleScaledValue*lengthscaleScale, periodScaledValue*periodScale]).double())
            kernelsParamsScales[k] = torch.tensor([lengthscaleScale, periodScale])
        else:
            raise ValueError("Invalid kernel type {:s} for latent {:d}".format(kernelType, k))
        kernels[k] = kernel
    answer = {"kernels": kernels, "kernelsParamsScales": kernelsParamsScales}
    return answer


def getVariationalMean0FromList(nLatents, nTrials, config):
    qMu0 = [[] for r in range(nLatents)]
    for k in range(nLatents):
        qMu0k0 = torch.tensor([float(str) for str in config["variational_params"]["qMu0Latent{:d}Trial0".format(k)][1:-1].split(", ")], dtype=torch.double)
        nIndPointsK = len(qMu0k0)
        qMu0[k] = torch.empty((nTrials, nIndPointsK, 1), dtype=torch.double)
        qMu0[k][0,:,0] = qMu0k0
        for r in range(1, nTrials):
            qMu0kr = torch.tensor([float(str) for str in config["variational_params"]["qMu0Latent{:d}Trial{:d}".format(k,r)][1:-1].split(", ")])
            qMu0[k][r,:,0] = qMu0kr
    return qMu0


def getIndPointsMeans(nTrials, nLatents, config):
    indPointsMeans = [[] for r in range(nTrials)]
    for r in range(nTrials):
        indPointsMeans[r] = [[] for k in range(nLatents)]
        for k in range(nLatents):
            indPointsMeansFN = config["indPoints_params"]["indPointsMeanFNLatent{:d}Trial{:d}".format(k,r)]
            indPointsMeans[r][k] = torch.reshape(torch.from_numpy(np.loadtxt(indPointsMeansFN)), (-1,1))
    return indPointsMeans


def getLatentsMeansFuncs(nLatents, nTrials, config):
    def getLatentMeanFunc(ampl, tau, freq, phase):
        mean = lambda t: ampl*torch.exp(-t/tau)*torch.sin(2*math.pi*freq*t + phase)
        return mean

    meansFuncs = [[] for r in range(nLatents)]
    for k in range(nLatents):
        ampl = float(config["latentMean_params"]["amplLatent{:d}".format(k)])
        tau = float(config["latentMean_params"]["tauLatent{:d}".format(k)])
        freq = float(config["latentMean_params"]["freqLatent{:d}".format(k)])
        phase = float(config["latentMean_params"]["phaseLatent{:d}".format(k)])
        meanFunc = getLatentMeanFunc(ampl=ampl, tau=tau, freq=freq, phase=phase)
        meansFuncs[k] = meanFunc
    return meansFuncs






import pdb
import torch

class KLDivergence:

    def __init__(self, indPointsLocsKMS, svPosteriorOnIndPoints):
        super(KLDivergence, self).__init__()
        self._indPointsLocsKMS = indPointsLocsKMS
        self._svPosteriorOnIndPoints = svPosteriorOnIndPoints

    def get_indPointsLocsKMS(self):
        return self._indPointsLocsKMS

    def get_svPosteriorOnIndPoints(self):
        return self._svPosteriorOnIndPoints

    def evalSumAcrossLatentsAndTrials(self):
        klDiv = 0
        qSigma = self._svPosteriorOnIndPoints.buildCov()
        nLatents = len(qSigma)
        for k in range(nLatents):
            klDivK = self._evalSumAcrossTrials(
                Kzz=self._indPointsLocsKMS.getKzz()[k],
                qMu=self._svPosteriorOnIndPoints.getMean()[k],
                qSigma=qSigma[k],
                latentIndex=k)
            klDiv += klDivK
        return klDiv

    def _evalSumAcrossTrials(self, Kzz, qMu, qSigma, latentIndex):
        # ESS \in nTrials x nInd x nInd
        ESS = qSigma + torch.matmul(qMu, qMu.permute(0,2,1))
        nTrials = qMu.shape[0]
        answer = 0
        for trialIndex in range(nTrials):
            _, logdetKzz = Kzz[trialIndex,:,:].slogdet() # O(n^3)
            _, logdetQSigma = qSigma[trialIndex,:,:].slogdet() # O(n^3)
            # traceTerm = torch.trace(torch.cholesky_solve(ESS[trialIndex,:,:],
            # KzzInv[trialIndex,:,:]))
            traceTerm = torch.trace(self._indPointsLocsKMS.solveForLatentAndTrial(ESS[trialIndex,:,:], latentIndex=latentIndex, trialIndex=trialIndex))
            trialKL = .5*(traceTerm+logdetKzz-logdetQSigma-ESS.shape[1])
            answer += trialKL
        return answer




import numpy as np
import torch
import svGPFA.stats.kernelsMatricesStore
import svGPFA.utils.miscUtils


def buildFloatListFromStringRep(stringRep):
    float_list = [float(str) for str in stringRep[1:-1].split(", ")]
    return float_list


def flatToHierarchicalOptimParams(flat_optim_params):
    hierarchical_optim_params = {
        "n_quad": flat_optim_params["n_quad"],
        "prior_cov_reg_param": flat_optim_params["prior_cov_reg_param"],
        "optim_method": flat_optim_params["optim_method"],
        "em_max_iter": flat_optim_params["em_max_iter"],
        "verbose": flat_optim_params["verbose"],
        # estep_
        "estep_estimate": flat_optim_params["estep_estimate"],
        "estep_optim_params": {
            "max_iter": flat_optim_params["estep_max_iter"],
            "lr": flat_optim_params["estep_lr"],
            "tolerance_grad": flat_optim_params["estep_tolerance_grad"],
            "tolerance_change": flat_optim_params["estep_tolerance_change"],
            "line_search_fn": flat_optim_params["estep_line_search_fn"],
        },
        # mstep_embedding_
        "mstep_embedding_estimate": flat_optim_params[
            "mstep_embedding_estimate"],
        "mstep_embedding_optim_params": {
            "max_iter": flat_optim_params["mstep_embedding_max_iter"],
            "lr": flat_optim_params["mstep_embedding_lr"],
            "tolerance_grad": flat_optim_params[
                "mstep_embedding_tolerance_grad"],
            "tolerance_change": flat_optim_params[
                "mstep_embedding_tolerance_change"],
            "line_search_fn": flat_optim_params[
                "mstep_embedding_line_search_fn"],
        },
        # mstep_kernels_
        "mstep_kernels_estimate": flat_optim_params["mstep_kernels_estimate"],
        "mstep_kernels_optim_params": {
            "max_iter": flat_optim_params["mstep_kernels_max_iter"],
            "lr": flat_optim_params["mstep_kernels_lr"],
            "tolerance_grad": flat_optim_params[
                "mstep_kernels_tolerance_grad"],
            "tolerance_change": flat_optim_params[
                "mstep_kernels_tolerance_change"],
            "line_search_fn": flat_optim_params[
                "mstep_kernels_line_search_fn"],
        },
        # mstep_indpointslocs_
        "mstep_indpointslocs_estimate": flat_optim_params[
            "mstep_indpointslocs_estimate"],
        "mstep_indpointslocs_optim_params": {
            "max_iter": flat_optim_params[
                "mstep_indpointslocs_max_iter"],
            "lr": flat_optim_params[
                "mstep_indpointslocs_lr"],
            "tolerance_grad": flat_optim_params[
                "mstep_indpointslocs_tolerance_grad"],
            "tolerance_change": flat_optim_params[
                "mstep_indpointslocs_tolerance_change"],
            "line_search_fn": flat_optim_params[
                "mstep_indpointslocs_line_search_fn"],
        },
    }
    return hierarchical_optim_params


def getOptimParams(dynamic_params_spec, config_file_params_spec,
                   default_params_spec=None,
                   optim_params_info=None, section_name="optim_params"):
    if optim_params_info is None:
        optim_params_info = getArgsInfo()["optim_params"]

    flat_optim_params = {}
    for param_name in optim_params_info:
        # param_conv_func = optim_params_info[param_name]
        param_value = getParam(section_name=section_name,
                               param_name=param_name,
                               dynamic_params_spec=dynamic_params_spec,
                               config_file_params_spec=config_file_params_spec,
                               default_params_spec=default_params_spec)
        flat_optim_params[param_name] = param_value
    hierarchical_optim_params = flatToHierarchicalOptimParams(
        flat_optim_params=flat_optim_params)
    return hierarchical_optim_params


def getDefaultParamsDict(n_neurons, n_trials, n_latents=3,
                         n_ind_points=None, common_n_ind_points=10, n_quad=200,
                         diag_var_cov0_value=1e-2, prior_cov_reg_param=1e-3,
                         lengthscale=1.0, em_max_iter=50):
    if n_ind_points is None:
        n_ind_points = [common_n_ind_points] * n_latents
    var_mean0 = [torch.zeros((n_trials, n_ind_points[k], 1),
                             dtype=torch.double)
                 for k in range(n_latents)]
    var_cov0 = [[] for r in range(n_latents)]
    for k in range(n_latents):
        var_cov0[k] = torch.empty((n_trials, n_ind_points[k], n_ind_points[k]),
                                  dtype=torch.double)
        for r in range(n_trials):
            var_cov0[k][r, :, :] = torch.eye(n_ind_points[k]) * \
                    diag_var_cov0_value

    params_dict = {
        "variational_params0": {
            "variational_mean0": var_mean0,
            "variational_cov0": var_cov0,
        },
        "embedding_params0": {
            "c0_distribution": "Normal",
            "c0_loc": 0.0,
            "c0_scale": 1.0,
            "d0_distribution": "Normal",
            "d0_loc": 0.0,
            "d0_scale": 1.0,
        },
        "kernels_params0": {
            "k_types": "exponentialQuadratic",
            "k_lengthscales0": lengthscale,
        },
        "ind_points_locs_params0": {
            "n_ind_points": n_ind_points,
            "ind_points_locs0_layout": "equidistant",
        },
        "optim_params": {
            "n_quad": n_quad,
            "prior_cov_reg_param": prior_cov_reg_param,
            "optim_method": "ecm",
            "em_max_iter": em_max_iter,
            "verbose": True,
            #
            "estep_estimate": True,
            "estep_max_iter": 20,
            "estep_lr": 1.0,
            "estep_tolerance_grad": 1e-7,
            "estep_tolerance_change": 1e-9,
            "estep_line_search_fn": "strong_wolfe",
            #
            "mstep_embedding_estimate": True,
            "mstep_embedding_max_iter": 20,
            "mstep_embedding_lr": 1.0,
            "mstep_embedding_tolerance_grad": 1e-7,
            "mstep_embedding_tolerance_change": 1e-9,
            "mstep_embedding_line_search_fn": "strong_wolfe",
            #
            "mstep_kernels_estimate": True,
            "mstep_kernels_max_iter": 20,
            "mstep_kernels_lr": 1.0,
            "mstep_kernels_tolerance_grad": 1e-7,
            "mstep_kernels_tolerance_change": 1e-9,
            "mstep_kernels_line_search_fn": "strong_wolfe",
            #
            "mstep_indpointslocs_estimate": True,
            "mstep_indpointslocs_max_iter": 20,
            "mstep_indpointslocs_lr": 1.0,
            "mstep_indpointslocs_tolerance_grad": 1e-7,
            "mstep_indpointslocs_tolerance_change": 1e-9,
            "mstep_indpointslocs_line_search_fn": "strong_wolfe"}
    }
    return params_dict


def strTo1DTensor(aString, dtype=np.float64, sep=","):
    if aString[0] in "[(" and aString[-1] in "])":
        aString = aString[1:-1]
    an_np_array = np.fromstring(aString, sep=sep, dtype=dtype)
    a_tensor = torch.from_numpy(an_np_array)
    return a_tensor


def strTo2DTensor(aString, dtype=torch.double):
    if dtype == torch.double:
        dtype_np = np.float64
    elif dtype == torch.int:
        dtype_np = np.int32
    else:
        raise ValueError(f"Invalid dtype={dtype}")

    an_np_matrix = np.matrix(aString, dtype=dtype_np)
    a_tensor = torch.from_numpy(an_np_matrix)
    return a_tensor


def strTo2DDoubleTensor(aString):
    return strTo2DTensor(aString=aString, dtype=torch.double)


def strTo2DIntTensor(aString):
    return strTo2DTensor(aString=aString, dtype=torch.int)


def strTo1DDoubleTensor(aString):
    return strTo1DTensor(aString=aString, dtype=np.float64)


def strTo1DIntTensor(aString):
    return strTo1DTensor(aString=aString, dtype=np.int32)


def getArgsInfo():
    args_info = {"model_structure_params": {
                     "n_latents": int,
                 },
                 "data_structure_params": {
                     "trials_start_time": float,
                     "trials_end_time": float,
                     "trials_start_times": strTo1DDoubleTensor,
                     "trials_end_times": strTo1DDoubleTensor
                 },
                 "variational_params0": {
                     "variational_means0": strTo1DDoubleTensor,
                     "variational_covs0": strTo2DDoubleTensor,
                     "variational_means0_filename": str,
                     "variational_covs0_filename": str,
                     "variational_mean0_constant_value": float,
                     "variational_cov0_diag_value": float,
                     "variational_mean0_filename_latent{:d}_trial{:d}": str,
                     "variational_cov0_filename_latent{:d}_trial{:d}": str,
                 },
                 "embedding_params0": {
                     "c0": strTo2DDoubleTensor,
                     "d0": strTo2DDoubleTensor,
                     "c0_filename": str,
                     "d0_filename": str,
                     "c0_distribution": str,
                     "c0_loc": float,
                     "c0_scale": float,
                     "c0_low": float,
                     "c0_high": float,
                     "c0_random_seed": int,
                     "d0_distribution": str,
                     "d0_loc": float,
                     "d0_scale": float,
                     "d0_low": float,
                     "d0_high": float,
                     "d0_random_seed": int,
                 },
                 "kernels_params0": {
                     "k_type": str,
                     "k_lengthscale0": float,
                     "k_period0": float,
                     "k_types": str,
                     "k_lengthscales0": float,
                     "k_period0s": float,
                     "k_type_latent{:d}": str,
                     "k_lengthscale0_latent{:d}": float,
                     "k_period0_latent{:d}": float,
                 },
                 "ind_points_locs_params0": {
                     "n_ind_points": strTo1DIntTensor,
                     "common_n_ind_points": int,
                     "ind_points_locs0": strTo1DDoubleTensor,
                     "ind_points_locs0_layout": str,
                     "ind_points_locs0_filename": str,
                     "ind_points_locs0_filename_latent{:d}_trial{:d}": str,
                 },
                 "optim_params": {
                     "n_quad": int,
                     "prior_cov_reg_param": float,
                     "optim_method": str,
                     "em_max_iter": int,
                     "verbose": bool,
                     #
                     "estep_estimate": bool,
                     "estep_max_iter": int,
                     "estep_lr": float,
                     "estep_tolerance_grad": float,
                     "estep_tolerance_change": float,
                     "estep_line_search_fn": str,
                     #
                     "mstep_embedding_estimate": bool,
                     "mstep_embedding_max_iter": int,
                     "mstep_embedding_lr": float,
                     "mstep_embedding_tolerance_grad": float,
                     "mstep_embedding_tolerance_change": float,
                     "mstep_embedding_line_search_fn": str,
                     #
                     "mstep_kernels_estimate": bool,
                     "mstep_kernels_max_iter": int,
                     "mstep_kernels_lr": float,
                     "mstep_kernels_tolerance_grad": float,
                     "mstep_kernels_tolerance_change": float,
                     "mstep_kernels_line_search_fn": str,
                     #
                     "mstep_indpointslocs_estimate": bool,
                     "mstep_indpointslocs_max_iter": int,
                     "mstep_indpointslocs_lr": float,
                     "mstep_indpointslocs_tolerance_grad": float,
                     "mstep_indpointslocs_tolerance_change": float,
                     "mstep_indpointslocs_line_search_fn": str,
                 }}
    return args_info


def getParamsDictFromArgs(n_latents, n_trials, args, args_info):
    # args is a non-hierarchical dictionary
    # args_info is a two-level-hierarchical dictionary
    params_dict = {}
    for key1 in args_info:
        params_dict[key1] = {}
        for key2 in args_info[key1]:
            conversion_func = args_info[key1][key2]
            if "_latent{:d}" in key2:
                for k in range(n_latents):
                    if "_trial{:d}" in key2:
                        for r in range(n_trials):
                            # latent{:d} should appear before _trial{:d} in the
                            # argument label
                            arg_name = key2.format(k, r)
                            if arg_name in args:
                                params_dict[key1][arg_name] = \
                                    conversion_func(args[arg_name])
                    else:
                        arg_name = key2.format(k)
                        if arg_name in args:
                            params_dict[key1][arg_name] = \
                                conversion_func(args[arg_name])
            else:
                arg_name = key2
                if arg_name in args:
                    params_dict[key1][arg_name] = conversion_func(
                        args[arg_name])
    return params_dict


def getParamsDictFromStringsDict(n_latents, n_trials, strings_dict, args_info):
    params_dict = {}
    for key1 in args_info:
        params_dict[key1] = {}
        for key2 in args_info[key1]:
            conversion_func = args_info[key1][key2]
            if "_latent{:d}" in key2:
                for k in range(n_latents):
                    if "_trial{:d}" in key2:
                        for r in range(n_trials):
                            # latent{:d} should appear before _trial{:d} in the
                            # argument label
                            arg_name = key2.format(k, r)
                            if key1 in strings_dict and \
                               arg_name in strings_dict[key1]:
                                params_dict[key1][arg_name] = \
                                    conversion_func(
                                        strings_dict[key1][arg_name])
                    else:
                        arg_name = key2.format(k)
                        if key1 in strings_dict and \
                           arg_name in strings_dict[key1]:
                            params_dict[key1][arg_name] = \
                                conversion_func(strings_dict[key1][arg_name])
            else:
                arg_name = key2
                if key1 in strings_dict and \
                   arg_name in strings_dict[key1]:
                    params_dict[key1][arg_name] = \
                        conversion_func(strings_dict[key1][arg_name])
    return params_dict


def getParamsAndKernelsTypes(n_neurons, n_trials, n_latents,
                             trials_start_times, trials_end_times,
                             default_params_spec=None,
                             config_file_params_spec=None,
                             dynamic_params_spec=None):
    """Builds initial, expected log likelihood and optimization parameters, as
    well as kernels types, from specifications given in
    ``default_params_spec``, ``config_file_params_spec`` and
    ``dynamic_params_spec``.

    :param n_neurons: number of neurons.
    :type  n_neurons: integer

    :param n_trials: number of trials.
    :type  n_trials: integer

    :param n_latents: number of latents.
    :type  n_latents: integer

    :param default_params_spec: default parameters specification formated as described in :ref:`params_specs` and usually obtained using  :func:`svGPFA.utils.initUtils.getDefaultParamsDict`,
    :type default_params_spec: list of dictionaries

    :param config_file_params_spec: parameters specification obtained from a configuration file, formated as described in :ref:`params_specs` and usually obtained using ``gcnu_common.utils.config_dict.GetDict`` followed by :func:`svGPFA.utils.initUtils.getParamsDictFromStringsDict`,

    :type config_file_params_spec: list of dictionaries

    :param dynamic_params_spec: parameters specification obtained from command line arguments, formated as described in :ref:`params_specs` and usually obtained using :func:`svGPFA.utils.initUtils.getParamsDictFromArgs`,

    :type dynamic_params_spec: list of dictionaries

    """

    n_quad = getParam(section_name="optim_params",
                      param_name="n_quad",
                      dynamic_params_spec=dynamic_params_spec,
                      config_file_params_spec=config_file_params_spec,
                      default_params_spec=default_params_spec)
    n_ind_points = getParam(section_name="ind_points_locs_params0",
                            param_name="n_ind_points",
                            dynamic_params_spec=dynamic_params_spec,
                            config_file_params_spec=config_file_params_spec,
                            default_params_spec=default_params_spec)
    if n_ind_points is None:
        common_n_ind_points = getParam(
            section_name="ind_points_locs_params0",
            param_name="common_n_ind_points",
            dynamic_params_spec=dynamic_params_spec,
            config_file_params_spec=config_file_params_spec,
            default_params_spec=default_params_spec)
        n_ind_points = [common_n_ind_points] * n_latents

    C0, d0 = getLinearEmbeddingParams0(
        n_neurons=n_neurons, n_latents=n_latents,
        dynamic_params_spec=dynamic_params_spec,
        config_file_params_spec=config_file_params_spec,
        default_params_spec=default_params_spec)

    leg_quad_points, leg_quad_weights = \
        svGPFA.utils.miscUtils.getLegQuadPointsAndWeights(
            n_quad=n_quad, trials_start_times=trials_start_times,
            trials_end_times=trials_end_times)

    kernels_params0, kernels_types = \
        getKernelsParams0AndTypes(
            n_latents=n_latents,
            dynamic_params_spec=dynamic_params_spec,
            config_file_params_spec=config_file_params_spec,
            default_params_spec=default_params_spec)

    ind_points_locs0 = getIndPointsLocs0(
        n_latents=n_latents, n_trials=n_trials,
        dynamic_params_spec=dynamic_params_spec,
        config_file_params_spec=config_file_params_spec,
        default_params_spec=default_params_spec,
        n_ind_points=n_ind_points,
        trials_start_times=trials_start_times,
        trials_end_times=trials_end_times,
    )

    var_mean0 = getVariationalMean0(
        n_latents=n_latents, n_trials=n_trials, n_ind_points=n_ind_points,
        dynamic_params_spec=dynamic_params_spec,
        config_file_params_spec=config_file_params_spec,
        default_params_spec=default_params_spec)

    var_cov0 = getVariationalCov0(
        n_latents=n_latents, n_trials=n_trials,
        dynamic_params_spec=dynamic_params_spec,
        config_file_params_spec=config_file_params_spec,
        default_params_spec=default_params_spec,
        n_ind_points=n_ind_points)
    var_cov0_chol = [svGPFA.utils.miscUtils.chol3D(var_cov0[k])
                     for k in range(n_latents)]
    var_cov0_chol_vecs = \
        svGPFA.utils.miscUtils.getVectorRepOfLowerTrianMatrices(
            lt_matrices=var_cov0_chol)

    optim_params = getOptimParams(
        dynamic_params_spec=dynamic_params_spec,
        config_file_params_spec=config_file_params_spec,
        default_params_spec=default_params_spec,
    )
    variational_params0 = {"mean": var_mean0,
                           "cholVecs": var_cov0_chol_vecs}
    kms_params0 = {"kernels_params0": kernels_params0,
                   "inducing_points_locs0": ind_points_locs0}
    posteriorLatentsParams0 = {"posterior_on_ind_points": variational_params0,
                               "kernels_matrices_store": kms_params0}
    embeddingParams0 = {"C0": C0, "d0": d0}
    initial_params = {"posterior_on_latents": posteriorLatentsParams0,
                      "embedding": embeddingParams0}
    quad_params = {"leg_quad_points": leg_quad_points,
                   "leg_quad_weights": leg_quad_weights}
    params = {"initial_params": initial_params,
              "ell_calculation_params": quad_params,
              "optim_params": optim_params}
    return params, kernels_types


def getParam(section_name, param_name,
             dynamic_params_spec=None,
             config_file_params_spec=None,
             default_params_spec=None):
    # dynamic_params_spec
    if dynamic_params_spec is not None and \
            section_name in dynamic_params_spec and \
            param_name in dynamic_params_spec[section_name]:
        param = dynamic_params_spec[section_name][param_name]
        print(f"Extracted dynamic_params_spec[{section_name}][{param_name}]="
              f"{param}")
    # config_file_params_spec
    elif config_file_params_spec is not None and \
            section_name in config_file_params_spec and \
            param_name in config_file_params_spec[section_name]:
        param = config_file_params_spec[section_name][param_name]
        print("Extracted "
              f"config_file_params_spec[{section_name}][{param_name}]="
              f"{param}")
    # default_params_spec
    elif default_params_spec is not None and \
            section_name in default_params_spec and \
            param_name in default_params_spec[section_name]:
        param = default_params_spec[section_name][param_name]
        print(f"Extracted default_params_spec[{section_name}][{param_name}]="
              f"{param}")
    else:
        param = None
    return param


def getLinearEmbeddingParams0(n_neurons, n_latents, dynamic_params_spec=None,
                              config_file_params_spec=None,
                              default_params_spec=None):
    C = getLinearEmbeddingParam0(
            param_label="c0", n_rows=n_neurons,
            n_cols=n_latents,
            dynamic_params_spec=dynamic_params_spec,
            config_file_params_spec=config_file_params_spec,
            default_params_spec=default_params_spec)
    d = getLinearEmbeddingParam0(
            param_label="d0", n_rows=n_neurons, n_cols=1,
            dynamic_params_spec=dynamic_params_spec,
            config_file_params_spec=config_file_params_spec,
            default_params_spec=default_params_spec)
    C = C.contiguous()
    d = d.contiguous()
    return C, d


def getLinearEmbeddingParam0(param_label, n_rows, n_cols,
                             dynamic_params_spec=None,
                             config_file_params_spec=None,
                             default_params_spec=None):
    if dynamic_params_spec is not None:
        param = getLinearEmbeddingParam0InDict(param_label=param_label,
                                               params_dict=dynamic_params_spec,
                                               params_dict_type="dynamic",
                                               n_rows=n_rows, n_cols=n_cols)
        if param is not None:
            return param

    if config_file_params_spec is not None:
        param = getLinearEmbeddingParam0InDict(
            param_label=param_label,
            params_dict=config_file_params_spec,
            params_dict_type="config_file",
            n_rows=n_rows, n_cols=n_cols)
        if param is not None:
            return param

    if default_params_spec is not None:
        param = getLinearEmbeddingParam0InDict(param_label=param_label,
                                               params_dict=default_params_spec,
                                               params_dict_type="default",
                                               n_rows=n_rows, n_cols=n_cols)
        if param is not None:
            return param

    raise ValueError("embedding_params_spec_params0 not found")


def getLinearEmbeddingParam0InDict(param_label, params_dict,
                                   params_dict_type, n_rows, n_cols,
                                   section_name="embedding_params0",
                                   delimiter=","):
    # binary
    if section_name in params_dict and \
       f"{param_label}" in params_dict[section_name]:
        param = params_dict[section_name][param_label]
        print(f"Extracted from {params_dict_type} {param_label}")
    # filename
    elif section_name in params_dict and \
            f"{param_label}_filename" in params_dict[section_name]:
        param_filename = params_dict[section_name][f"{param_label}_filename"]
        param_np = np.genfromtxt(param_filename, delimiter=delimiter)
        param = torch.from_numpy(param_np).type(torch.double)
        print(f"Extracted from {params_dict_type} {param_label}_filename")
    # random
    elif section_name in params_dict and \
            f"{param_label}_distribution" in params_dict[section_name]:
        param_distribution = \
            params_dict[section_name][f"{param_label}_distribution"]
        if f"{param_label}_random_seed" in params_dict[section_name]:
            param_random_seed = \
                params_dict[section_name][f"{param_label}_random_seed"]
        else:
            param_random_seed = None
        # If param_random_seed was specified for replicability
        if param_random_seed is not None:
            torch.random.manual_seed(param_random_seed)
        if param_distribution == "Normal":
            if f"{param_label}_loc" in params_dict[section_name] and \
               f"{param_label}_scale" in params_dict[section_name]:
                param_loc = params_dict[section_name][f"{param_label}_loc"]
                param_scale = params_dict[section_name][f"{param_label}_scale"]
                param = torch.distributions.normal.Normal(
                    param_loc, param_scale).sample(
                        sample_shape=[n_rows, n_cols]).type(torch.double)
                print(f"Extracted from {params_dict_type} "
                      f"{param_label}_distribution={param_distribution}, "
                      f"{param_label}_loc={param_loc}, "
                      f"{param_label}_scale={param_scale}, "
                      f"{param_label}_random_seed={param_random_seed}")

            else:
                raise ValueError(
                    f"For a {param_distribution} distribution "
                    f"{param_label}_loc and {param_label}_scale need to be "
                    "provided")
        elif param_distribution == "Uniform":
            if f"{param_label}_low" in params_dict[section_name] and \
               f"{param_label}_high" in params_dict[section_name]:
                param_low = params_dict[section_name][f"{param_label}_low"]
                param_high = params_dict[section_name][f"{param_label}_high"]
                param = torch.distributions.uniform.Uniform(
                    low=param_low, high=param_high).sample(
                        sample_shape=[n_rows, n_cols]).type(torch.double)
                print(f"Extracted from {params_dict_type} "
                    f"{param_label}_distribution={param_distribution}, "
                    f"{param_label}_low={param_low}, "
                    f"{param_label}_high={param_high}, "
                    f"{param_label}_random_seed={param_random_seed}")
            else:
                raise ValueError(
                    f"For a {param_distribution} distribution "
                    f"{param_label}_low and {param_label}_high need to be "
                    "provided")
        else:
            raise ValueError(
                f"Invalid param_distribution={param_distribution}")
        # If param_random_seed was specified for replicability
        if param_random_seed is not None:
            torch.random.seed()
    else:
        param = None

    return param


def getKernelsParams0AndTypes(n_latents, dynamic_params_spec=None,
                              config_file_params_spec=None,
                              default_params_spec=None):
    if dynamic_params_spec is not None:
        params0, kernels_types = getKernelsParams0AndTypesInDict(
            n_latents=n_latents, params_dict=dynamic_params_spec,
            params_dict_type="dynamic")
        if params0 is not None and kernels_types is not None:
            return params0, kernels_types

    if config_file_params_spec is not None:
        params0, kernels_types = getKernelsParams0AndTypesInDict(
            n_latents=n_latents, params_dict=config_file_params_spec,
            params_dict_type="config_file")
        if params0 is not None and kernels_types is not None:
            return params0, kernels_types

    if default_params_spec is not None:
        params0, kernels_types = getKernelsParams0AndTypesInDict(
            n_latents=n_latents, params_dict=default_params_spec,
            params_dict_type="default")
        if params0 is not None and kernels_types is not None:
            return params0, kernels_types

    raise ValueError("kernels parameters not found")


def getKernelsParams0AndTypesInDict(n_latents, params_dict, params_dict_type,
                                    section_name="kernels_params0"):
    # binary format
    if section_name in params_dict and \
       "k_types" in params_dict[section_name] and \
       "k_params0" in params_dict[section_name]:
        kernels_types = params_dict[section_name]["k_types"]
        params0 = params_dict[section_name]["k_params0"]
        print(f"Extracted from {params_dict_type} k_types={kernels_types} and "
              f"k_params0={params0}")
    # short format
    elif section_name in params_dict and \
            "k_types" in params_dict[section_name]:
        if params_dict[section_name]["k_types"] == "exponentialQuadratic":
            kernels_types = ["exponentialQuadratic" for k in range(n_latents)]
            if "k_lengthscales0" in params_dict[section_name]:
                lengthscales0 = \
                    float(params_dict[section_name]["k_lengthscales0"])
            else:
                raise ValueError("If k_type=exponentialQuadratic is specified "
                                 f"in {params_dict_type}, then "
                                 "k_lengthscales0 should also be specified in "
                                 f"{params_dict_type}")
            params0 = [torch.DoubleTensor([lengthscales0])
                       for k in range(n_latents)]
            print(f"Extracted from {params_dict_type} "
                  "k_type=exponentialQuadratic and "
                  f"k_lengthsales0={lengthscales0}")
        elif params_dict[section_name]["k_types"] == "periodic":
            kernels_types = ["periodic" for k in range(n_latents)]
            if "k_lengthscales0" in params_dict[section_name]:
                lengthscales0 = \
                    float(params_dict[section_name]["k_lengthscales0"])
            else:
                raise ValueError("If k_types=periodic is specified "
                                 f"in {params_dict_type}, then "
                                 "k_lengthscales0 should also be specified in "
                                 f"{params_dict_type}")
            if "k_periods0" in params_dict[section_name]:
                periods0 = float(params_dict[section_name]["k_periods0"])
            else:
                raise ValueError("If k_types=periodic is specified "
                                 f"in {params_dict_type}, then k_periods0 "
                                 "should also be specified in "
                                 f"{params_dict_type}")
            params0 = [torch.DoubleTensor([lengthscales0, periods0])
                       for k in range(n_latents)]
            print(f"Extracted from {params_dict_type} "
                  f"k_types=periodic, k_lengthsales0={lengthscales0} "
                  f"and  k_periods={periods0}")
    # long format
    elif section_name in params_dict and \
            "k_type_latent0" in params_dict[section_name]:
        kernels_types = []
        params0 = []
        for k in range(n_latents):
            if params_dict[section_name][f"k_type_latent{k}"] == \
                    "exponentialQuadratic":
                kernels_types.append("exponentialQuadratic")
                if f"k_lengthscale0_latent{k}" in params_dict[section_name]:
                    lengthscale0 = float(
                        params_dict[section_name][f"k_lengthscale0_latent{k}"])
                else:
                    raise ValueError("If k_type=exponentialQuadratic is "
                                     "specified in {params_dict_type}, "
                                     f"then k_lengthscale0_latent{k} "
                                     "should also be specified in "
                                     f"{params_dict_type}")
                params0.append(torch.DoubleTensor([lengthscale0]))
                print(f"Extracted from  {params_dict_type} "
                      f"k_type_latent{k}=exponentialQuadratic and "
                      f"k_lengthsale0_latent0{k}={lengthscale0}")
            elif params_dict[section_name][f"k_type_latent{k}"] == "periodic":
                kernels_types.append("periodic")
                if "k_lengthscale0_latent{k}" in params_dict[section_name]:
                    lengthscale0 = float(
                        params_dict[section_name][f"k_lengthscale0_latent{k}"])
                else:
                    raise ValueError("If k_type=periodic is "
                                     "specified in {params_dict_type}, "
                                     f"then k_lengthscale0_latent{k} "
                                     "should also be specified in "
                                     f"{params_dict_type}")
                if "k_period0_latent{k}" in params_dict[section_name]:
                    period0 = float(
                        params_dict[section_name][f"k_period0_latent{k}"])
                else:
                    raise ValueError("If k_type=periodic is "
                                     f"specified in {params_dict_type}, "
                                     f"then k_period0_latent{k} "
                                     "should also be specified in "
                                     f"{params_dict_type}")
                params0.append(torch.DoubleTensor([lengthscale0, period0]))
                print(f"Extracted from {params_dict_type} "
                      f"k_type_latent{k}=periodic, "
                      f"k_lengthsale0_latent{k}={lengthscale0} and "
                      f"k_period0_latent{k}={period0}")
            else:
                raise RuntimeError("Invalid k_type_latent{:d}={:s}".format(
                    k, params_dict[section_name][f"k_type_latent{k}"]))
    else:
        params0 = None
        kernels_types = None
    return params0, kernels_types


def getIndPointsLocs0(n_latents, n_trials,
                      dynamic_params_spec=None, config_file_params_spec=None,
                      default_params_spec=None, n_ind_points=None,
                      trials_start_times=None, trials_end_times=None):
    if dynamic_params_spec is not None:
        param = getIndPointsLocs0InDict(n_latents=n_latents, n_trials=n_trials,
                                        params_dict=dynamic_params_spec,
                                        params_dict_type="dynamic",
                                        n_ind_points=n_ind_points,
                                        trials_start_times=trials_start_times,
                                        trials_end_times=trials_end_times)
        if param is not None:
            return param

    if config_file_params_spec is not None:
        param = getIndPointsLocs0InDict(n_latents=n_latents, n_trials=n_trials,
                                        params_dict=config_file_params_spec,
                                        params_dict_type="config_file",
                                        n_ind_points=n_ind_points,
                                        trials_start_times=trials_start_times,
                                        trials_end_times=trials_end_times)
        if param is not None:
            return param

    if default_params_spec is not None:
        param = getIndPointsLocs0InDict(n_latents=n_latents, n_trials=n_trials,
                                        params_dict=default_params_spec,
                                        params_dict_type="default",
                                        n_ind_points=n_ind_points,
                                        trials_start_times=trials_start_times,
                                        trials_end_times=trials_end_times)
        if param is not None:
            return param

    raise ValueError("ind_points_locs0 not found")


def getIndPointsLocs0InDict(n_latents, n_trials, params_dict, params_dict_type,
                            n_ind_points, trials_start_times, trials_end_times,
                            section_name="ind_points_locs_params0"):
    # binary
    if section_name in params_dict and \
       "ind_points_locs0" in params_dict[section_name]:
        ind_points_locs0 = params_dict[section_name]["ind_points_locs0"]
        print(f"Extracted from {params_dict_type} ind_points_locs ")
    # filename: same inducing points across all latents and trials
    elif section_name in params_dict and \
            "ind_points_locs0_filename" in params_dict[section_name]:
        ind_points_locs0_filename = \
            params_dict[section_name]["ind_points_locs0_filename"]
        ind_points_locs0 = getSameAcrossLatentsAndTrialsIndPointsLocs0(
            n_latents=n_latents, n_trials=n_trials,
            ind_points_locs0_filename=ind_points_locs0_filename)
        print(f"Extracted from {params_dict_type} "
              f"ind_points_locs0_filename={ind_points_locs0_filename}")
    # filename: different inducing points across all latents and trials
    elif section_name in params_dict and \
            "ind_points_locs0_filename_latent0_trial0" in \
            params_dict[section_name]:
        ind_points_locs0 = getDiffAcrossLatentsAndTrialsIndPointsLocs0(
            n_latents=n_latents, n_trials=n_trials, params_dict=params_dict,
            params_dict_type=params_dict_type)
    # layout
    elif section_name in params_dict and \
            "ind_points_locs0_layout" in params_dict[section_name] and \
            n_ind_points is not None and  \
            trials_start_times is not None and \
            trials_end_times is not None:
        layout = params_dict[section_name]["ind_points_locs0_layout"]
        print(f"Extracted from {params_dict_type} "
              f"ind_points_locs0_layout={layout}")
        if layout == "equidistant":
            ind_points_locs0 = buildEquidistantIndPointsLocs0(
                n_latents=n_latents, n_trials=n_trials,
                n_ind_points=n_ind_points,
                trials_start_times=trials_start_times,
                trials_end_times=trials_end_times)
        elif layout == "uniform":
            ind_points_locs0 = buildUniformIndPointsLocs0(
                n_latents=n_latents, n_trials=n_trials,
                n_ind_points=n_ind_points,
                trials_start_times=trials_start_times,
                trials_end_times=trials_end_times)
        else:
            raise RuntimeError(f"Invalid ind_points_locs0_layout={layout}")
    else:
        ind_points_locs0 = None

    return ind_points_locs0


def getSameAcrossLatentsAndTrialsIndPointsLocs0(
        n_latents, n_trials, ind_points_locs0_filename, delimiter=","):
    Z0_np = np.genfromtxt(ind_points_locs0_filename,
                          delimiter=delimiter).flatten()
    Z0 = torch.from_numpy(Z0_np)
    Z0s = [[] for k in range(n_latents)]
    nIndPointsForLatent = len(Z0)
    for k in range(n_latents):
        Z0s[k] = torch.empty((n_trials, nIndPointsForLatent, 1),
                             dtype=torch.double)
        Z0s[k][:, :, 0] = Z0
    return Z0s


def getDiffAcrossLatentsAndTrialsIndPointsLocs0(
        n_latents, n_trials, params_dict, params_dict_type,
        section_name="ind_points_locs_params0",
        item_name_pattern="ind_points_locs0_filename_latent{:d}_trial{:d}",
        delimiter=","):
    Z0 = [[] for k in range(n_latents)]
    for k in range(n_latents):
        item_name = item_name_pattern.format(k, 0)
        ind_points_locs0_filename = params_dict[section_name][item_name]
        print(f"Extracted from {params_dict_type} "
              f"{item_name}={ind_points_locs0_filename}")
        Z0_k_r0_np = np.genfromtxt(ind_points_locs0_filename,
                                   delimiter=delimiter)
        Z0_k_r0 = torch.from_numpy(Z0_k_r0_np).flatten()
        nIndPointsForLatent = len(Z0_k_r0)
        Z0[k] = torch.empty((n_trials, nIndPointsForLatent, 1),
                            dtype=torch.double)
        Z0[k][0, :, 0] = Z0_k_r0
        for r in range(1, n_trials):
            item_name = item_name_pattern.format(k, r)
            ind_points_locs0_filename = params_dict[section_name][item_name]
            print(f"Extracted from {params_dict_type} "
                  f"{item_name}={ind_points_locs0_filename}")
            Z0_k_r_np = np.genfromtxt(ind_points_locs0_filename,
                                      delimiter=delimiter)
            Z0_k_r = torch.from_numpy(Z0_k_r_np).flatten()
            Z0[k][r, :, 0] = Z0_k_r
    return Z0


def buildEquidistantIndPointsLocs0(n_latents, n_trials, n_ind_points,
                                   trials_start_times, trials_end_times):
    Z0s = [[] for k in range(n_latents)]
    for k in range(n_latents):
        Z0s[k] = torch.empty((n_trials, n_ind_points[k], 1),
                             dtype=torch.double)
        for r in range(n_trials):
            Z0 = torch.linspace(trials_start_times[r], trials_end_times[r],
                                n_ind_points[k])
            Z0s[k][r, :, 0] = Z0
    return Z0s


def buildUniformIndPointsLocs0(n_latents, n_trials, n_ind_points,
                               trials_start_times, trials_end_times):
    Z0s = [[] for k in range(n_latents)]
    for k in range(n_latents):
        Z0s[k] = torch.empty((n_trials, n_ind_points[k], 1),
                             dtype=torch.double)
        for r in range(n_trials):
            Z0 = trials_start_times[r] + \
                 torch.rand(n_ind_points[k]) * \
                 (trials_end_times[r]-trials_start_times[r])
            Z0_sorted, _ = Z0.sort()
            Z0s[k][r, :, 0] = Z0_sorted
    return Z0s


def getVariationalMean0(n_latents, n_trials, n_ind_points=None,
                        dynamic_params_spec=None, config_file_params_spec=None,
                        default_params_spec=None):
    if dynamic_params_spec is not None:
        param = getVariationalMean0InDict(n_latents=n_latents,
                                          n_trials=n_trials,
                                          n_ind_points=n_ind_points,
                                          params_dict=dynamic_params_spec,
                                          params_dict_type="dynamic")
        if param is not None:
            return param

    if config_file_params_spec is not None:
        param = getVariationalMean0InDict(n_latents=n_latents,
                                          n_trials=n_trials,
                                          n_ind_points=n_ind_points,
                                          params_dict=config_file_params_spec,
                                          params_dict_type="config_file")
        if param is not None:
            return param

    if default_params_spec is not None:
        param = getVariationalMean0InDict(n_latents=n_latents,
                                          n_trials=n_trials,
                                          n_ind_points=n_ind_points,
                                          params_dict=default_params_spec,
                                          params_dict_type="default")
        if param is not None:
            return param

    raise ValueError("variational_mean0 not found")


def getVariationalMean0InDict(
        n_latents, n_trials, n_ind_points,
        params_dict, params_dict_type,
        section_name="variational_params0",
        binary_item_name="variational_mean0",
        common_filename_item_name="variational_means0_filename",
        different_filename_item_name_pattern="variational_mean0_filename_latent{:d}_trial{:d}",
        constant_value_item_name="variational_mean0_constant_value",
        delimiter=","):
    # binary
    if section_name in params_dict and \
       binary_item_name in params_dict[section_name]:
        variational_mean0 = params_dict[section_name][binary_item_name]
        print(f"Extracted from {params_dict_type} {binary_item_name}")
    # constant_value
    elif section_name in params_dict and \
            constant_value_item_name in params_dict[section_name]:
        constant_value = params_dict[section_name][constant_value_item_name]
        # make sure all latents have the same number of inducing points
        for k in range(1, len(n_ind_points)):
            assert(n_ind_points[0] == n_ind_points[k])
        a_variational_mean0 = torch.ones(n_ind_points[0],
                                         dtype=torch.double) * constant_value
        print(f"Extracted from {params_dict_type} "
              f"{constant_value_item_name}={constant_value}")
        variational_mean0 = getSameAcrossLatentsAndTrialsVariationalMean0(
            n_latents=n_latents, n_trials=n_trials,
            a_variational_mean0=a_variational_mean0)
    # variational_means_filename
    elif section_name in params_dict and \
            common_filename_item_name in params_dict[section_name]:
        variational_mean0_filename = \
            params_dict[section_name][common_filename_item_name]
        print(f"Extracted from {params_dict_type} "
              f"{common_filename_item_name}={variational_mean0_filename}")
        a_variational_mean0_np = np.genfromtxt(variational_mean0_filename,
                                               delimiter=delimiter)
        a_variational_mean0 = \
            torch.from_numpy(a_variational_mean0_np).flatten()
        variational_mean0 = getSameAcrossLatentsAndTrialsVariationalMean0(
            n_latents=n_latents, n_trials=n_trials,
            a_variational_mean0=a_variational_mean0)
    # variational_means_filename latent k trial r
    elif section_name in params_dict and \
            different_filename_item_name_pattern.format(0, 0) \
            in params_dict[section_name]:
        variational_mean0 = getDiffAcrossLatentsAndTrialsVariationalMean0(
            n_latents=n_latents, n_trials=n_trials, params_dict=params_dict,
            params_dict_type=params_dict_type,
            section_name=section_name,
            item_name_pattern=different_filename_item_name_pattern)
    else:
        variational_mean0 = None
    return variational_mean0


def getSameAcrossLatentsAndTrialsVariationalMean0(n_latents, n_trials,
                                                  a_variational_mean0):
    n_ind_points = len(a_variational_mean0)
    variational_mean0 = [[] for r in range(n_latents)]
    for k in range(n_latents):
        variational_mean0[k] = torch.empty((n_trials, n_ind_points, 1),
                                           dtype=torch.double)
        variational_mean0[k][:, :, 0] = a_variational_mean0
    return variational_mean0


def getDiffAcrossLatentsAndTrialsVariationalMean0(
        n_latents, n_trials, params_dict, params_dict_type,
        section_name="variational_params0",
        item_name_pattern="variational_mean_latent{:d}_trial{:d}_filename",
        delimiter=","):
    variational_mean0 = [[] for r in range(n_latents)]
    for k in range(n_latents):
        variational_mean0_filename = \
            params_dict[section_name][item_name_pattern.format(k, 0)]
        print(f"Extracted from {params_dict_type} "
              f"{item_name_pattern.format(k, 0)}={variational_mean0_filename}")
        variational_mean0_k0_np = np.genfromtxt(variational_mean0_filename,
                                                delimiter=delimiter)
        variational_mean0_k0 = \
            torch.from_numpy(variational_mean0_k0_np).flatten()
        nIndPointsK = len(variational_mean0_k0)
        variational_mean0[k] = torch.empty((n_trials, nIndPointsK, 1),
                                           dtype=torch.double)
        variational_mean0[k][0, :, 0] = variational_mean0_k0
        for r in range(1, n_trials):
            variational_mean0_filename = \
                params_dict[section_name][item_name_pattern.format(k, r)]
            print(f"Extracted from {params_dict_type} "
                  f"{item_name_pattern.format(k, r)}="
                  f"{variational_mean0_filename}")
            variational_mean0_kr = np.genfromtxt(variational_mean0_filename,
                                                 delimiter=delimiter)
            variational_mean0_kr = \
                torch.from_numpy(variational_mean0_kr).flatten()
            variational_mean0[k][r, :, 0] = variational_mean0_kr
    return variational_mean0


def getVariationalCov0(n_latents, n_trials,
                       dynamic_params_spec=None, config_file_params_spec=None,
                       default_params_spec=None, n_ind_points=None):
    if dynamic_params_spec is not None:
        param = getVariationalCov0InDict(n_latents=n_latents,
                                         n_trials=n_trials,
                                         n_ind_points=n_ind_points,
                                         params_dict=dynamic_params_spec,
                                         params_dict_type="dynamic")
        if param is not None:
            return param

    if config_file_params_spec is not None:
        param = getVariationalCov0InDict(n_latents=n_latents,
                                         n_trials=n_trials,
                                         n_ind_points=n_ind_points,
                                         params_dict=config_file_params_spec,
                                         params_dict_type="config_file")
        if param is not None:
            return param

    if default_params_spec is not None:
        param = getVariationalCov0InDict(n_latents=n_latents,
                                         n_trials=n_trials,
                                         n_ind_points=n_ind_points,
                                         params_dict=default_params_spec,
                                         params_dict_type="default")
        if param is not None:
            return param

    raise ValueError("variationalCov0 not found")


def getVariationalCov0InDict(
        n_latents, n_trials, params_dict, params_dict_type,
        n_ind_points=None, section_name="variational_params0",
        binary_item_name="variational_cov0",
        common_filename_item_name="variational_covs0_filename",
        different_filename_item_name_pattern="variational_cov0_filename_latent{:d}_trial{:d}",
        diag_value_item_name="variational_cov0_diag_value",
        delimiter=","):
    # binary variational mean and cov
    if section_name in params_dict and \
       binary_item_name in params_dict[section_name]:
        variational_cov0 = params_dict[section_name][binary_item_name]
        print(f"Extracted from {params_dict_type} {binary_item_name}")
    # diag_value
    elif section_name in params_dict and \
            diag_value_item_name in params_dict[section_name]:
        diag_value = params_dict[section_name][diag_value_item_name]
        print(f"Extracted from {params_dict_type} "
              f"{diag_value_item_name}={diag_value}")
        # make sure all latents have the same number of inducing points
        for k in range(1, len(n_ind_points)):
            assert(n_ind_points[0] == n_ind_points[k])
        a_variational_cov0 = diag_value * torch.eye(n_ind_points[0],
                                                    dtype=torch.double)
        variational_cov0 = getSameAcrossLatentsAndTrialsVariationalCov0(
            n_latents=n_latents, n_trials=n_trials,
            a_variational_cov0=a_variational_cov0,
            section_name=section_name)
    # common_filename
    elif section_name in params_dict and \
            common_filename_item_name in params_dict[section_name]:
        variational_cov0_filename = \
            params_dict[section_name][common_filename_item_name]
        print(f"Extracted from {params_dict_type} "
              f"{common_filename_item_name}={variational_cov0_filename}")
        a_variational_cov0_np = np.genfromtxt(variational_cov0_filename,
                                              delimiter=delimiter)
        a_variational_cov0 = torch.from_numpy(a_variational_cov0_np)
        variational_cov0 = getSameAcrossLatentsAndTrialsVariationalCov0(
            n_latents=n_latents, n_trials=n_trials,
            a_variational_cov0=a_variational_cov0,
            section_name=section_name)
    # variational_cov0_latent{:d}_trial{:d}_filename
    elif section_name in params_dict and \
            different_filename_item_name_pattern.format(0, 0) \
            in params_dict[section_name]:
        variational_cov0 = getDiffAcrossLatentsAndTrialsVariationalCov0(
            n_latents=n_latents, n_trials=n_trials,
            params_dict=params_dict, params_dict_type=params_dict_type)
    else:
        variational_cov0 = None
    return variational_cov0


def getSameAcrossLatentsAndTrialsVariationalCov0(
        n_latents, n_trials, a_variational_cov0,
        section_name="variational_params0"):
    variational_cov0 = [[] for r in range(n_latents)]
    n_ind_points = a_variational_cov0.shape[0]
    for k in range(n_latents):
        variational_cov0[k] = torch.empty((n_trials, n_ind_points,
                                           n_ind_points), dtype=torch.double)
        variational_cov0[k][:, :, :] = a_variational_cov0
    return variational_cov0


def getDiffAcrossLatentsAndTrialsVariationalCov0(
        n_latents, n_trials, params_dict, params_dict_type,
        section_name="variational_params0",
        item_name_pattern="variational_cov0_filename_latent{:d}_trial{:d}",
        delimiter=","):
    variational_cov0 = [[] for r in range(n_latents)]
    for k in range(n_latents):
        item_name = item_name_pattern.format(k, 0)
        variational_cov_filename = params_dict[section_name][item_name]
        print(f"Extracted from config[{section_name}] "
              f"{item_name}={variational_cov_filename}")
        variational_cov0_k0_np = np.genfromtxt(variational_cov_filename,
                                               delimiter=delimiter)
        variational_cov0_k0 = torch.from_numpy(variational_cov0_k0_np)
        nIndPointsK = variational_cov0_k0.shape[0]
        variational_cov0[k] = torch.empty((n_trials, nIndPointsK, nIndPointsK),
                                          dtype=torch.double)
        variational_cov0[k][0, :, :] = variational_cov0_k0
        for r in range(1, n_trials):
            item_name = item_name_pattern.format(k, r)
            variational_cov_filename = params_dict[section_name][item_name]
            print(f"Extracted from config[{section_name}] "
                  f"{item_name}={variational_cov_filename}")
            variational_cov0_kr_np = np.genfromtxt(variational_cov_filename,
                                                   delimiter=delimiter)
            variational_cov0_kr = torch.from_numpy(variational_cov0_kr_np)
            variational_cov0[k][r, :, :] = variational_cov0_kr
    return variational_cov0


def getUniformIndPointsMeans(n_trials, n_latents, nIndPointsPerLatent,
                             min=-1, max=1):
    ind_points_means = [[] for r in range(n_trials)]
    for r in range(n_trials):
        ind_points_means[r] = [[] for k in range(n_latents)]
        for k in range(n_latents):
            ind_points_means[r][k] = \
                torch.rand(nIndPointsPerLatent[k], 1)*(max - min)+min
    return ind_points_means


def getConstantIndPointsMeans(constantValue, n_trials, n_latents,
                              nIndPointsPerLatent):
    ind_points_means = [[] for r in range(n_trials)]
    for r in range(n_trials):
        ind_points_means[r] = [[] for k in range(n_latents)]
        for k in range(n_latents):
            ind_points_means[r][k] = \
                    constantValue*torch.ones(nIndPointsPerLatent[k], 1,
                                             dtype=torch.double)
    return ind_points_means


def getKzzChol0(kernels, kernelsParams0, ind_points_locs0, epsilon):
    ind_points_locs_kms = svGPFA.stats.kernelsMatricesStore.IndPointsLocsKMS()
    ind_points_locs_kms.setKernels(kernels=kernels)
    ind_points_locs_kms.setKernelsParams(kernelsParams=kernelsParams0)
    ind_points_locs_kms.setIndPointsLocs(ind_points_locs=ind_points_locs0)
    ind_points_locs_kms.setEpsilon(epsilon=epsilon)
    ind_points_locs_kms.buildKernelsMatrices()
    KzzChol0 = ind_points_locs_kms.getKzzChol()
    return KzzChol0


def getScaledIdentityQSigma0(scale, n_trials, nIndPointsPerLatent):
    nLatent = len(nIndPointsPerLatent)
    qSigma0 = [[None] for k in range(nLatent)]

    for k in range(nLatent):
        qSigma0[k] = torch.empty((n_trials, nIndPointsPerLatent[k],
                                  nIndPointsPerLatent[k]), dtype=torch.double)
        for r in range(n_trials):
            qSigma0[k][r, :, :] = scale*torch.eye(nIndPointsPerLatent[k],
                                                  dtype=torch.double)
    return qSigma0


def getSVPosteriorOnIndPointsParams0(nIndPointsPerLatent, n_latents, n_trials,
                                     scale):
    qMu0 = [[] for k in range(n_latents)]
    qSVec0 = [[] for k in range(n_latents)]
    qSDiag0 = [[] for k in range(n_latents)]
    for k in range(n_latents):
        qMu0[k] = torch.zeros(n_trials, nIndPointsPerLatent[k], 1,
                              dtype=torch.double)
        qSVec0[k] = scale*torch.eye(nIndPointsPerLatent[k], 1,
                                    dtype=torch.double).repeat(n_trials, 1, 1)
        qSDiag0[k] = scale*torch.ones(nIndPointsPerLatent[k], 1,
                                      dtype=torch.double).repeat(
                                          n_trials, 1, 1)
    return qMu0, qSVec0, qSDiag0


def getKernelsParams0(kernels, noiseSTD):
    n_latents = len(kernels)
    kernelsParams0 = [kernels[k].getParams() for k in range(n_latents)]
    if noiseSTD > 0.0:
        kernelsParams0 = [kernelsParams0[0] +
                          noiseSTD*torch.randn(len(kernelsParams0[k]))
                          for k in range(n_latents)]
    return kernelsParams0


def getKernelsScaledParams0(kernels, noiseSTD):
    n_latents = len(kernels)
    kernelsParams0 = [kernels[k].getScaledParams() for k in range(n_latents)]
    if noiseSTD > 0.0:
        kernelsParams0 = [kernelsParams0[0] +
                          noiseSTD*torch.randn(len(kernelsParams0[k]))
                          for k in range(n_latents)]
    return kernelsParams0




import pdb
from abc import ABC, abstractmethod
import math
import torch

class Kernel(ABC):

    @abstractmethod
    def buildKernelMatrix(self, X1, X2=None):
        pass

    @abstractmethod
    def buildKernelMatrixDiag(self, X):
        pass

    def getParams(self):
        return self._params

    def setParams(self, params):
        self._params = params

    @abstractmethod
    def getNamedParams(self):
        pass

class ExponentialQuadraticKernel(Kernel):

    def __init__(self, scale=1.0, lengthscaleScale=1.0, dtype=torch.double):
        self._scale = torch.tensor(scale, dtype=dtype)
        self._lengthscaleScale = lengthscaleScale

    def buildKernelMatrix(self, X1, X2=None):
        scale, lengthscale = self._getAllParams()
        if not hasattr(self, "_lengthscaleScale"):
            self._lengthscaleScale = 1.0
        lengthscale = lengthscale/self._lengthscaleScale

        if X2 is None:
            X2 = X1
        if X1.ndim==3:
            distance = (X1-X2.transpose(1, 2))**2
        else:
            distance = (X1.reshape(-1,1)-X2.reshape(1,-1))**2
        covMatrix = scale**2*torch.exp(-.5*distance/lengthscale**2)
        return covMatrix

    def buildKernelMatrixDiag(self, X):
        scale, lengthscale = self._getAllParams()
        covMatrixDiag = scale**2*torch.ones(X.shape, dtype=X.dtype, device=X.device)
        return covMatrixDiag

    def _getAllParams(self):
        scale = self._scale
        lengthscale = self._params[0]
        return scale, lengthscale

    def getScaledParams(self):
        scaledParams = torch.tensor([self._params[0]/self._lengthscaleScale])
        return scaledParams

    def getNamedParams(self):
        scale, lengthscale = self._getAllParams()
        answer = {"scale": scale, "lengthscale": lengthscale}
        return answer

class PeriodicKernel(Kernel):
    def __init__(self, scale=1.0, lengthscaleScale=1.0, periodScale=1.0, dtype=torch.double):
        self._scale = torch.tensor(scale, dtype=dtype)
        self._lengthscaleScale = lengthscaleScale
        self._periodScale = periodScale

    def buildKernelMatrix(self, X1, X2=None):
        scale, lengthscale, period = self._getAllParams()
        lengthscale = lengthscale/self._lengthscaleScale
        period = period/self._periodScale
        if X2 is None:
            X2 = X1
        if X1.ndim==3:
            sDistance = X1-X2.transpose(1, 2)
        else:
            sDistance = X1.reshape(-1,1)-X2.reshape(1,-1)
        rr = math.pi*sDistance/period
        covMatrix = scale**2*torch.exp(-2*torch.sin(rr)**2/lengthscale**2)
        return covMatrix

    def buildKernelMatrixDiag(self, X):
        scale, lengthscale, period = self._getAllParams()
        covMatrixDiag = scale**2*torch.ones(X.shape, dtype=X.dtype, device=X.device)
        return covMatrixDiag

    def _getAllParams(self):
        scale = self._scale
        lengthscale = self._params[0]
        period = self._params[1]

        return scale, lengthscale, period

    def getScaledParams(self):
        scaledParams = torch.tensor([self._params[0]/self._lengthscaleScale,
                                     self._params[1]/self._periodScale])
        return scaledParams

    def getNamedParams(self):
        scale, lengthscale, period = self._getAllParams()
        answer = {"scale": scale, "lengthscale": lengthscale, "period": period}
        return answer





import svGPFA.stats.kernelsMatricesStore
import svGPFA.stats.svPosteriorOnIndPoints
import svGPFA.stats.svPosteriorOnLatents
import svGPFA.stats.svEmbedding
import svGPFA.stats.expectedLogLikelihood
import svGPFA.stats.klDivergence
import svGPFA.stats.svLowerBound

#:
PointProcess = 0
#:
Poisson = 1
#:
Gaussian = 2


#:
LinearEmbedding = 100

#:
ExponentialLink = 1000
#:
NonExponentialLink = 1001

#:
kernelMatrixInvChol = 10000
#:
kernelMatrixInvPInv = 10001

#:
indPointsCovRank1PlusDiag = 100000
#:
indPointsCovChol = 100001


class SVGPFAModelFactory:

    @staticmethod
    def buildModelPyTorch(kernels, conditionalDist=PointProcess,
                          linkFunction=ExponentialLink,
                          embeddingType=LinearEmbedding,
                          kernelMatrixInvMethod=kernelMatrixInvChol,
                          indPointsCovRep=indPointsCovChol):
        """Creates an svGPFA model. :meth:`svGPFA.stats.svLowerBound.SVLowerBound.setInitialDataAndParams` should be invoked before using the created model as argument to :meth:`svGPFA.stats.svEM.SVEM.maximize`.

        :param kernels: list of kernels (:mod:`svGPFA.stats.kernel`) to be used in the model
        :type kernels: list of instances from subclass of (:class:`svGPFA.stats.kernel.Kernel`)
        :param conditionalDist: likelihood distribution (e.g., svGPFA.stats.SVGPFAModelFactory.PointProcess or svGPFA.stats.SVGPFAModelFactory.Gaussian)
        :type conditionalDist: int
        :param embeddingType: type of embedding (e.g., svGPFA.stats.SVGPFAModelFactory.LinearEmbedding)
        :type embeddingType: int

        :return: an unitialized model. Parameters and data need to be set (by calling :meth:`svGPFA.stats.svLowerBound.SVLowerBound.setParamsAndData`) before invoking :meth:`svGPFA.stats.svEM.SVEM.maximize`.
        :rtype: an instance of :class:`svGPFA.stats.svLowerBound.SVLowerBound`.

        """

        if conditionalDist == PointProcess:
            if embeddingType == LinearEmbedding:
                if linkFunction == ExponentialLink:
                    if indPointsCovRep == indPointsCovChol:
                        qU = svGPFA.stats.svPosteriorOnIndPoints.\
                                SVPosteriorOnIndPointsChol()
                    elif indPointsCovRep == indPointsCovRank1PlusDiag:
                        qU = svGPFA.stats.svPosteriorOnIndPoints.\
                                SVPosteriorOnIndPointsRank1PlusDiag()
                    else:
                        raise ValueError("Invalid indPointsCovRep")
                    if kernelMatrixInvMethod == kernelMatrixInvChol:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                                IndPointsLocsKMS_Chol()
                    elif kernelMatrixInvMethod == kernelMatrixInvPInv:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                                IndPointsLocsKMS_PInv()
                    else:
                        raise ValueError("Invalid kernelMatrixInvMethod")
                    indPointsLocsAndAllTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
                    indPointsLocsAndAssocTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAssocTimesKMS()
                    qKAllTimes = svGPFA.stats.svPosteriorOnLatents.\
                        SVPosteriorOnLatentsAllTimes(
                            svPosteriorOnIndPoints=qU,
                            indPointsLocsKMS=indPointsLocsKMS,
                            indPointsLocsAndTimesKMS=indPointsLocsAndAllTimesKMS)
                    qKAssocTimes = svGPFA.stats.svPosteriorOnLatents.\
                        SVPosteriorOnLatentsAssocTimes(
                            svPosteriorOnIndPoints=qU,
                            indPointsLocsKMS=indPointsLocsKMS,
                            indPointsLocsAndTimesKMS=indPointsLocsAndAssocTimesKMS)
                    qHAllTimes = svGPFA.stats.svEmbedding.\
                        LinearSVEmbeddingAllTimes(
                            svPosteriorOnLatents=qKAllTimes)
                    qHAssocTimes = svGPFA.stats.svEmbedding.\
                        LinearSVEmbeddingAssocTimes(
                            svPosteriorOnLatents=qKAssocTimes)
                    eLL = svGPFA.stats.expectedLogLikelihood.PointProcessELLExpLink(
                        svEmbeddingAllTimes=qHAllTimes,
                        svEmbeddingAssocTimes=qHAssocTimes)
                    klDiv = svGPFA.stats.klDivergence.KLDivergence(
                        indPointsLocsKMS=indPointsLocsKMS,
                        svPosteriorOnIndPoints=qU)
                    svlb = svGPFA.stats.svLowerBound.SVLowerBound(eLL=eLL,
                                                                  klDiv=klDiv)
                    svlb.setKernels(kernels=kernels)
                else:
                    raise ValueError("Invalid linkFunction={:s}".
                                     format(repr(linkFunction)))
            else:
                raise ValueError("Invalid embeddingType={:s}".
                                 format(repr(embeddingType)))
        elif conditionalDist == Poisson:
            if embeddingType == LinearEmbedding:
                if linkFunction == ExponentialLink:
                    qU = svGPFA.stats.svPosteriorOnIndPoints.\
                            SVPosteriorOnIndPoints()
                    if kernelMatrixInvMethod == kernelMatrixInvChol:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                                IndPointsLocsKMS_Chol()
                    elif kernelMatrixInvMethod == kernelMatrixInvPInv:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                                IndPointsLocsKMS_Pinv()
                    else:
                        raise ValueError("Invalid kernelMatrixInvMethod")
                    indPointsLocsAndAllTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
                    qKAllTimes = svGPFA.stats.svPosteriorOnLatents.\
                        SVPosteriorOnLatentsAllTimes(
                            svPosteriorOnIndPoints=qU,
                            indPointsLocsKMS=indPointsLocsKMS,
                            indPointsLocsAndTimesKMS=indPointsLocsAndAllTimesKMS)
                    qHAllTimes = svGPFA.stats.svEmbedding.\
                        LinearSVEmbeddingAllTimes(
                            svPosteriorOnLatents=qKAllTimes)
                    eLL = svGPFA.stats.expectedLogLikelihood.PoissonELLExpLink(
                        svEmbeddingAllTimes=qHAllTimes)
                    klDiv = svGPFA.stats.klDivergence.KLDivergence(
                        indPointsLocsKMS=indPointsLocsKMS,
                        svPosteriorOnIndPoints=qU)
                    svlb = svGPFA.stats.svLowerBound.SVLowerBound(eLL=eLL,
                                                                  klDiv=klDiv)
                    svlb.setKernels(kernels=kernels)
        else:
            raise ValueError("Invalid conditionalDist=%s"%
                             repr(conditionalDist))

        return svlb

    @staticmethod
    def buildModelSciPy(conditionalDist, linkFunction, embeddingType, kernels,
                        kernelMatrixInvMethod, indPointsCovRep):

        if conditionalDist == PointProcess:
            if embeddingType == LinearEmbedding:
                if linkFunction == ExponentialLink:
                    if indPointsCovRep == indPointsCovChol:
                        qU = svGPFA.stats.svPosteriorOnIndPoints.\
                            SVPosteriorOnIndPointsCholWithGettersAndSetters()
                    elif indPointsCovRep == indPointsCovRank1PlusDiag:
                        qU = svGPFA.stats.svPosteriorOnIndPoints.\
                            SVPosteriorOnIndPointsRank1PlusDiagWithGettersAndSetters()
                    else:
                        raise ValueError("Invalid indPointsCovRep")
                    if kernelMatrixInvMethod == kernelMatrixInvChol:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                            IndPointsLocsKMS_CholWithGettersAndSetters()
                    elif kernelMatrixInvMethod == kernelMatrixInvPInv:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                            IndPointsLocsKMS_PInvWithGettersAndSetters()
                    else:
                        raise ValueError("Invalid kernelMatrixInvMethod")
                    indPointsLocsAndAllTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
                    indPointsLocsAndAssocTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAssocTimesKMS()
                    qKAllTimes = svGPFA.stats.svPosteriorOnLatents.\
                        SVPosteriorOnLatentsAllTimes(
                            svPosteriorOnIndPoints=qU,
                            indPointsLocsKMS=indPointsLocsKMS,
                            indPointsLocsAndTimesKMS=indPointsLocsAndAllTimesKMS)
                    qKAssocTimes = svGPFA.stats.svPosteriorOnLatents.\
                            SVPosteriorOnLatentsAssocTimes(
                                svPosteriorOnIndPoints=qU,
                                indPointsLocsKMS=indPointsLocsKMS,
                                indPointsLocsAndTimesKMS=indPointsLocsAndAssocTimesKMS)
                    qHAllTimes = svGPFA.stats.svEmbedding.\
                            LinearSVEmbeddingAllTimesWithParamsGettersAndSetters(
                                svPosteriorOnLatents=qKAllTimes)
                    qHAssocTimes = svGPFA.stats.svEmbedding.LinearSVEmbeddingAssocTimes(
                        svPosteriorOnLatents=qKAssocTimes)
                    eLL = svGPFA.stats.expectedLogLikelihood.PointProcessELLExpLink(
                        svEmbeddingAllTimes=qHAllTimes,
                        svEmbeddingAssocTimes=qHAssocTimes)
                    klDiv = svGPFA.stats.klDivergence.KLDivergence(
                        indPointsLocsKMS=indPointsLocsKMS,
                        svPosteriorOnIndPoints=qU)
                    svlb = svGPFA.stats.svLowerBound.\
                        SVLowerBoundWithParamsGettersAndSetters(eLL=eLL,
                                                                klDiv=klDiv)
                    svlb.setKernels(kernels=kernels)
                else:
                    raise ValueError("Invalid linkFunction={:s}".
                                     format(repr(linkFunction)))
            else:
                raise ValueError("Invalid embeddingType={:s}".
                                 format(repr(embeddingType)))
        elif conditionalDist == Poisson:
            if embeddingType == LinearEmbedding:
                if linkFunction == ExponentialLink:
                    qU = svGPFA.stats.svPosteriorOnIndPoints.\
                        SVPosteriorOnIndPoints()
                    if kernelMatrixInvMethod == kernelMatrixInvChol:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                            IndPointsLocsKMS_Chol()
                    elif kernelMatrixInvMethod == kernelMatrixInvPInv:
                        indPointsLocsKMS = svGPFA.stats.kernelsMatricesStore.\
                            IndPointsLocsKMS_Pinv()
                    else:
                        raise ValueError("Invalid kernelMatrixInvMethod")
                    indPointsLocsAndAllTimesKMS = svGPFA.stats.\
                        kernelsMatricesStore.IndPointsLocsAndAllTimesKMS()
                    qKAllTimes = svGPFA.stats.svPosteriorOnLatents.\
                        SVPosteriorOnLatentsAllTimes(
                            svPosteriorOnIndPoints=qU,
                            indPointsLocsKMS=indPointsLocsKMS,
                            indPointsLocsAndTimesKMS=indPointsLocsAndAllTimesKMS)
                    qHAllTimes = svGPFA.stats.svEmbedding.\
                        LinearSVEmbeddingAllTimes(
                            svPosteriorOnLatents=qKAllTimes)
                    eLL = svGPFA.stats.expectedLogLikelihood.PoissonELLExpLink(
                        svEmbeddingAllTimes=qHAllTimes)
                    klDiv = svGPFA.stats.klDivergence.KLDivergence(
                        indPointsLocsKMS=indPointsLocsKMS,
                        svPosteriorOnIndPoints=qU)
                    svlb = svGPFA.stats.svLowerBound.SVLowerBound(
                        eLL=eLL, klDiv=klDiv)
                    svlb.setKernels(kernels=kernels)
        else:
            raise ValueError("Invalid conditionalDist={:s}".
                             format(repr(conditionalDist)))

        return svlb




import pdb
import sys
import io
import time
import math
import pickle
import numpy as np
import torch
import traceback

class TerminationInfo:
    def __init__(self, message):
        self._message = message

    @property
    def message(self):
        return self._message

class ErrorTerminationInfo(TerminationInfo):
    def __init__(self, message, error):
        super().__init__(message=message)
        self._error = error

    @property
    def error(self):
        return self._error

class SVLBFGS:

    def maximize(self, model, optimParams, method="EM", getIterationModelParamsFn=None,
                 logLock=None, logStreamFN=None,
                 lowerBoundLock=None, lowerBoundStreamFN=None,
                 latentsTimes=None, latentsLock=None, latentsStreamFN=None,
                 verbose=True, out=sys.stdout,
                 savePartial=False,
                 savePartialFilenamePattern="00000000_{:s}_estimatedModel.pickle",
                ):

        if latentsStreamFN is not None and latentsTimes is None:
            raise RuntimeError("Please specify latentsTime if you want to save latents")

        iter = 0
        if savePartial:
            savePartialFilename = savePartialFilenamePattern.format("initial")
            resultsToSave = {"model": model}
            with open(savePartialFilename, "wb") as f: pickle.dump(resultsToSave, f)
        lowerBound0 = model.eval()
        lowerBoundHist = [lowerBound0.item()]
        elapsedTimeHist = [0.0]
        startTime = time.time()

        if lowerBoundLock is not None and lowerBoundStreamFN is not None and not lowerBoundLock.is_locked():
            lowerBoundLock.lock()
            with open(lowerBoundStreamFN, 'wb') as f:
                np.save(f, np.array(lowerBoundHist))
            lowerBoundLock.unlock()

        if latentsLock is not None and latentsStreamFN is not None and not latentsLock.is_locked():
            latentsLock.lock()
            muK, varK = model.predictLatents(newTimes=latentsTimes)

            with open(latentsStreamFN, 'wb') as f:
                np.savez(f, iteration=iter, times=latentsTimes.detach().numpy(), muK=muK.detach().numpy(), varK=varK.detach().numpy())
            lowerBoundLock.unlock()
        iter += 1
        logStream = io.StringIO()
        if getIterationModelParamsFn is not None:
            initialModelsParams = getIterationModelParamsFn(model=model)
            iterationsModelParams = torch.empty((optimParams["em_max_iter"]+1, len(initialModelsParams)), dtype=torch.double)
            iterationsModelParams[0,:] = initialModelsParams
        else:
            iterationsModelParams = None
        maxRes = {"lowerBound": lowerBound0}
        while iter<optimParams["em_max_iter"]:
            message = "Iteration {:02d}, start: {:f}\n".format(iter, maxRes["lowerBound"])
            if verbose:
                out.write(message)
            self._writeToLockedLog(
                message=message,
                logLock=logLock,
                logStream=logStream,
                logStreamFN=logStreamFN
            )
            maxRes = self._optimizeAllParams(model=model,
                                             optimParams=optimParams["LBFGS_optim_params"])
            message = "Iteration {:02d}, end: {:f}, niter: {:d}, nfeval: {:d}\n".format(iter, maxRes["lowerBound"], maxRes["niter"], maxRes["nfeval"])
            if verbose:
                out.write(message)
            self._writeToLockedLog(
                message=message,
                logLock=logLock,
                logStream=logStream,
                logStreamFN=logStreamFN
            )
            if savePartial:
                savePartialFilename = savePartialFilenamePattern.format("{:s}{:03d}".format(step, iter))
                resultsToSave = {"model": model}
                with open(savePartialFilename, "wb") as f: pickle.dump(resultsToSave, f)
            if getIterationModelParamsFn is not None:
                iterationsModelParams[iter+1,:] = getIterationModelParamsFn(model=model)
            elapsedTimeHist.append(time.time()-startTime)
            lowerBoundHist.append(maxRes["lowerBound"].item())

            if lowerBoundLock is not None and lowerBoundStreamFN is not None and not lowerBoundLock.is_locked():
                lowerBoundLock.lock()
                with open(lowerBoundStreamFN, 'wb') as f:
                    np.save(f, np.array(lowerBoundHist))
                lowerBoundLock.unlock()

            if latentsLock is not None and latentsStreamFN is not None and not latentsLock.is_locked():
                latentsLock.lock()
                muK, varK = model.predictLatents(newTimes=latentsTimes)

                with open(latentsStreamFN, 'wb') as f:
                    np.savez(f, iteration=iter, times=latentsTimes.detach().numpy(), muK=muK.detach().numpy(), varK=varK.detach().numpy())
                lowerBoundLock.unlock()

            iter += 1
        terminationInfo = TerminationInfo("Maximum number of iterations ({:d}) reached".format(optimParams["em_max_iter"]))
        return lowerBoundHist, elapsedTimeHist, terminationInfo, iterationsModelParams

    def _eStep(self, model, optimParams):
        x = model.getSVPosteriorOnIndPointsParams()
        evalFunc = model.eval
        optimizer = torch.optim.LBFGS(x, **optimParams)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _mStepEmbedding(self, model, optimParams):
        x = model.getSVEmbeddingParams()
        # pdb.set_trace()
        # x = [i.contiguous() for i in x]
        svPosteriorOnLatentsStats = model.computeSVPosteriorOnLatentsStats()
        evalFunc = lambda: model.evalELLSumAcrossTrialsAndNeurons(svPosteriorOnLatentsStats=svPosteriorOnLatentsStats)
        # evalFunc = model.eval
        optimizer = torch.optim.LBFGS(x, **optimParams)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _mStepKernels(self, model, optimParams):
        x = model.getKernelsParams()
        def evalFunc():
            model.buildKernelsMatrices()
            answer = model.eval()
            return answer
        optimizer = torch.optim.LBFGS(x, **optimParams)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        print("Kernel params:", x)
        return answer

    def _mStepIndPointsLocs(self, model, optimParams):
        x = model.getIndPointsLocs()
        def evalFunc():
            model.buildKernelsMatrices()
            answer = model.eval()
            return answer
        optimizer = torch.optim.LBFGS(x, **optimParams)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _optimizeAllParams(self, model, optimParams):
        x = model.getSVPosteriorOnIndPointsParams() + \
            model.getSVEmbeddingParams() + \
            model.getKernelsParams() + \
            model.getIndPointsLocs()
        def evalFunc():
            model.buildKernelsMatrices()
            answer = model.eval()
            # print("Lower bound: {:.02f}".format(answer))
            return answer
        optimizer = torch.optim.LBFGS(x, **optimParams)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _setupAndMaximizeStep(self, x, evalFunc, optimizer):
        for i in range(len(x)):
            x[i].requires_grad = True
        maxRes = self._maximizeStep(evalFunc=evalFunc, optimizer=optimizer)
        for i in range(len(x)):
            x[i].requires_grad = False
        return maxRes

    def _maximizeStep(self, evalFunc, optimizer):
        def closure():
            optimizer.zero_grad()
            curEval = -evalFunc()
            curEval.backward(retain_graph=True)
            return curEval
        optimizer.step(closure)
        lowerBound = evalFunc()
        stateOneEpoch = optimizer.state[optimizer._params[0]]
        nfeval = stateOneEpoch["func_evals"]
        niter = stateOneEpoch["n_iter"]
        return {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter}

    def _writeToLockedLog(self, message, logLock, logStream, logStreamFN):
        logStream.write(message)
        if logLock is not None and not logLock.is_locked():
            logLock.lock()
            with open(logStreamFN, 'a') as f:
                f.write(logStream.getvalue())
            logLock.unlock()
            logStream.truncate(0)
            logStream.seek(0)

